/* $Workfile: T2p5x4.c $       $Revision: 82 $      $Modtime: 3/08/07 4:01p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    T2P5X4.C

    A simple example console application demonstrating the use the following
    devices:
    
      OTX_LDEVICE_TONE_DUAL_DETECTOR, 
      OTX_LDEVICE_TONE_DTMF_DETECTOR, 
      OTX_LDEVICE_TONE_MF_DETECTOR, 
      OTX_LDEVICE_TONE_SINEWAVE_GENERATOR,
      OTX_LDEVICE_TONE_DTMF_DIALER, 
      OTX_LDEVICE_TONE_MF_DIALER 
    
   The applications works with the following board combinations:

      Thor-2-PCI / Thor-8-PCI + Vidar-5x4-ASM,

    or OTX PCI Plus series boards,

      Gimle-16-PCI-Plus / Thor-2-PCI-Plus  / Thor-4-PCI-Plus / Thor-8-PCI-Plus + Vidar-5x4-ASM-CST,
      Gimle-16-PCI-Plus / Thor-2-PCI-Plus  / Thor-4-PCI-Plus / Thor-8-PCI-Plus + Vidar-5x4-ASM-PRO,
      Gimle-16-PCI-Plus / Thor-2-PCI-Plus  / Thor-4-PCI-Plus / Thor-8-PCI-Plus + Vidar-5x4-ASM-EX,
      Gimle-16-PCI-Plus / Thor-2-PCI-Plus  / Thor-4-PCI-Plus / Thor-8-PCI-Plus + Vidar-55x4-ASM,

    or OTX PCI Express series boards,

      Thor-2-PCI-Express + Vidar-5x4-ASM-CST,
      Thor-2-PCI-Express + Vidar-5x4-ASM-PRO,
      Thor-2-PCI-Express + Vidar-5x4-ASM-EX,
      Thor-2-PCI-Express + Vidar-55x4-ASM,

    or Thor-2-PCMCIA series boards,

      Thor-2-PCMCIA, 
      Thor-2-PCMCIA-Plus, 
      Thor-2-PCMCIA-Ex, 
      Thor-2-PCMCIA-Pro,
      Thor-2-PCMCIA-CST

    or Sleipnir-1-PCI-Plus

    or Thor-2-ExpressCard

    The sample application provides a simple console interface which utilizes
    simple keystrokes as commands.

    It configures 2 E1 Lines and expects them to be looped with an external
    loop cable. 

    2 Handsets are connected (only applicable for OTX boards equipped with a Codec) to 
    the codecs and a through connection is demonstrated through the Loop Cable.

    It also creates various tone device on the DSP, which are
    used to detect and generate tones.

    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#if !defined(_WIN32_WCE)
#include <assert.h>
#include <time.h>
#endif  // #if !defined(_WIN32_WCE)

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxErr.h"
#include "OtxBrd.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxTone.h"
#include "OtxDsp.h"
#include "OtxHdlc.h"
#include "OtxT1E1.h"
#include "OtxData.h"

#define DEMO_TEST_MF_DETECTOR 1

#define DEMO_DEFAULT_RX_THRESHOLD 500     // in mV

//#define DEMO_DUAL_TONE_DETECTOR 1   // Define this if the generic Dual Tone detector should be used instead of the DTMF detector

#define DEMO_TASK_REF 1L               // Task Reference for a non-blocking call

#define DEMO_DEFAULT_LOG_FILE "T2p5x4.log"
#define DEMO_DEFAULT_BOARD_NO 0

#define DEMO_ASM0_HIGHWAY    10
#define DEMO_ASM1_HIGHWAY    11
#define DEMO_ASM2_HIGHWAY    12
#define DEMO_ASM3_HIGHWAY    13
#define DEMO_CODEC_HIGHWAY   14

typedef struct DemoEnvT {
  char           *pszInputStr;
  OTX_BOOL        bExitOnError;
  OTX_INT32       nBoardType;
  char            cwd[OTX_MAX_PATH];
  char            logFileName[OTX_MAX_PATH];
  OtxDataFormatE  law;
  OtxT1E1LiModeE  eT1E1Mode;       // T1 or E1 mode
  OtxDspSpmE      eProgramModule;  // Standard Program Module to be loaded in the DSP (from the Lib)
  char           *pszDspOutFile;   // Optionally load the DSP program direct from an OUT file (this file name)
  int             boardNo;
  int             coDsp;
  int             coLi;
  OtxMfToneE      eMfToneType;  // Type of MF Tone to detect
  OTX_BOOL        bUseDspDebugger;
  OTX_BOOL        bLiLoop;

  OTX_BOOL               bReceiveEqualizer;
  OtxT1E1TermModeE       eTermMode;
  OTX_BOOL               bMonitorMode;
  OTX_BOOL               bAmplifier;
  OTX_INT32              nRxThreshold;

} DemoEnvT;


/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoOpenPhysicalDevices(void);
static void DemoConfigureE1T1Devices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoCreateLogicalDevices(void);
static void DemoConnectDevices(void);
static void DemoInitializeLogicalDevices(void);
static void DemoEnableLogicalDevices(void);
static void DemoInitT1E1ConfigData(OUT OtxE1ConfigOptionsS *pE1CfgData, OUT OtxT1ConfigOptionsS *pT1CfgData);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);
static void DemoConnectCodecSpeakerSource(OTX_INT32 ts);
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoHelpPanel(void);
static void DemoLogf(char *formatStr, ...);
static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);

static OTX_INT32 DemoProcessUserCommand(char commandChar);


/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;
OTX_BOOL    g_bContinousHdlcSendingActive = OTX_FALSE;
char *g_sqNumber0 = "0123456789ABCD*#";
OTX_INT32   g_nCodecTs = 0;  // Dsp Highway timeslot to connect the Codec to
OTX_UINT32  g_nDspHighwaySpeed;  // DSP Highway speed in kpbs

struct {
  OTX_HANDLE hThorController0;
  OTX_HANDLE hQuadE1T1[2]; // Only used for Thor-4-PCI-Plus, Thor-8-PCI(-Plus); not for Thor-2-PCI and Thor-2-PCMCIA
  OTX_HANDLE hVidarAsmController0;
  OTX_HANDLE hLi[2];
  OTX_HANDLE hCodec[4];
  OTX_HANDLE hDsp[4];
  OTX_HANDLE hTss;
} g_hPhysDevs;


struct {
  OTX_HANDLE hDspLiDtmfDialer[4];
  OTX_HANDLE hLiDspDtmfDetector[4];
  OTX_HANDLE hDspLiSineWaveSender[4];
  OTX_HANDLE hLiDspSineWaveReceiver[4];
  OTX_HANDLE hTssCodec[4]; 
  OTX_HANDLE hLiIn[2];
  OTX_HANDLE hLiOut[2];
} g_hPipes;


struct {
  OTX_HANDLE hHdlcSender[2];
  OTX_HANDLE hHdlcReceiver[2];
  OTX_HANDLE hDtmfDetector[4];
  OTX_HANDLE hDtmfDialler[4];
  OTX_HANDLE hMfDialler[4];
  OTX_HANDLE hSineWaveReceiver[4]; // actually a DTMF detector
  OTX_HANDLE hSineWaveSender[4];
} g_hLogicDevs;


struct {
  OtxE1ConfigOptionsS e1CfgData;
  OtxT1ConfigOptionsS t1CfgData;
} g_T1E1CfgData;


struct {
  OTX_HANDLE hHandle;
  OTX_HANDLE hFromHsPipe;
  OTX_HANDLE hToHsPipe;
} g_sqDataConverter[1];


/*******************************************************************************
*******************************************************************************/


/*******************************************************************************

  @func

  Create the event queue to be used by the driver to inform the application
  about events, such as line status changes, or detected DTMF digits.

*******************************************************************************/
void DemoCreateEventQueue(void)
{
  OTX_RESULT result;

  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);						

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(g_hEventQueue, result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC,0); DemoCheckRc(g_hEventQueue, result);
}



/*******************************************************************************

  @func
    Open the Physical Devices on the Thor board

  @comm
    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.

*******************************************************************************/
void DemoOpenPhysicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;
  OTX_INT32  nDsp;
  OTX_INT32  nQuadE1T1; // Only used for Thor-4-PCI-Plus, Thor-8-PCI(-Plus)

  if ((g_demoEnv.nBoardType == OTX_DEVICE_THOR_PCMCIA) && (g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2_EXTMEM)) {
    DemoLogf("The OTX_DSP_SPM_OTXSPM2_EXTMEM program is not supported on a OTX_DEVICE_THOR_PCMCIA board\n");
    exit(1);
  }

  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_PCMCIA:
    DemoLogf("Opening OTX_DEVICE_THOR_PCMCIA board\n");
    break;
  case OTX_DEVICE_THOR_PCI:
    DemoLogf("Opening OTX_DEVICE_THOR_PCI board\n");
    break;
  case OTX_DEVICE_THOR_8_PCI:
    DemoLogf("Opening Thor-8-PCI board\n");
    break;
  case OTX_DEVICE_THOR_2_PCI_PLUS:
    DemoLogf("Opening Thor-2-PCI-Plus board\n");
    break;
  case OTX_DEVICE_THOR_2_PCI_EXPRESS:
    DemoLogf("Opening Thor-2-PCI-Express board\n");
    break;
  case OTX_DEVICE_THOR_8_PCI_PLUS:
    DemoLogf("Opening Thor-8-PCI-Plus board\n");
    break;
  case OTX_DEVICE_THOR_4_PCI_PLUS:
    DemoLogf("Opening Thor-4-PCI-Plus board\n");
    break;
  case OTX_DEVICE_SLEIPNIR_1_PCI:
    DemoLogf("Opening Sleipnir-1-PCI board\n");
    break;
  case OTX_DEVICE_GIMLE_16_PCI_PLUS:
    DemoLogf("Opening Gimle-16-PCI-Plus board\n");
    break;
  case OTX_DEVICE_THOR_2_EXPRESSCARD:
    DemoLogf("Opening Thor-2-ExpressCard board\n");
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }
  result = OtxDrvOpenPhysicalDevice(0, (OtxPhysicalDeviceTypeE)g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hThorController0); DemoCheckRc(g_hPhysDevs.hThorController0, result);

  if ( (g_demoEnv.nBoardType == OTX_DEVICE_THOR_PCMCIA) || (g_demoEnv.nBoardType == OTX_DEVICE_SLEIPNIR_1_PCI)  || (g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_EXPRESSCARD) ) {
    // do nothing, because they don't have OTX_DEVICE_TSS
  } else {
    DemoLogf("Open OTX_DEVICE_TSS\n");
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_TSS, 0, 0, g_hEventQueue, &g_hPhysDevs.hTss); DemoCheckRc(g_hPhysDevs.hThorController0, result);
  }

  // Get handles to T1/E1 line interface devices #0 and #1
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_PCMCIA:
  case OTX_DEVICE_THOR_PCI:
  case OTX_DEVICE_THOR_2_PCI_PLUS:
  case OTX_DEVICE_THOR_2_PCI_EXPRESS:
  case OTX_DEVICE_SLEIPNIR_1_PCI:
    // On these boards the E1T1 devices are hosted by the board itself. Get handles to T1/E1 line interface devices #0 and #1
    for (i = 0; i < g_demoEnv.coLi; i++) {
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d\n", i);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    }
    break;
  case OTX_DEVICE_THOR_8_PCI_PLUS:
  case OTX_DEVICE_THOR_8_PCI:
    {
      // On the Thor-8-PCI and Thor-8-PCI-Plus board the E1T1 devices are hosted by a quad E1T1 device. Open those two devices.
      for (nQuadE1T1 = 0; nQuadE1T1 < 2; nQuadE1T1++) {
        DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      }
      
      // Get handles to T1/E1 line interface devices 
      for (i = 0; i < g_demoEnv.coLi; i++) {
        nQuadE1T1 = i/4;
        DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
      }
    }
    break;
  case OTX_DEVICE_THOR_4_PCI_PLUS:
    {
      // On the Thor-4-PCI-Plus board the E1T1 devices are hosted by a quad E1T1 device. 
      for (nQuadE1T1 = 0; nQuadE1T1 < 1; nQuadE1T1++) {
        DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      }
      
      // Get handles to T1/E1 line interface devices 
      for (i = 0; i < g_demoEnv.coLi; i++) {
        nQuadE1T1 = i/4;
        DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
      }
    }
    break;

  case OTX_DEVICE_GIMLE_16_PCI_PLUS:
    {
      // On the Gimle-16-PCI-Plus board the E1T1 devices are hosted by 4 quad E1T1 devices. Open those four devices.
      for (nQuadE1T1 = 0; nQuadE1T1 < 4; nQuadE1T1++) {
        DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      }
      
      // Get handles to T1/E1 line interface devices 
      for (i = 0; i < g_demoEnv.coLi; i++) {
        nQuadE1T1 = i/4;
        DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
      }
    }
    break;

  case OTX_DEVICE_THOR_2_EXPRESSCARD:
    {
      // On the Thor-2-ExpressCard board the E1T1 devices are hosted by a Dual E1T1 device. 
      for (nQuadE1T1 = 0; nQuadE1T1 < 1; nQuadE1T1++) {
        DemoLogf("Open OTX_DEVICE_DUAL_T1E1 #%d\n", nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_DUAL_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      }
      
      // Get handles to T1/E1 line interface devices 
      for (i = 0; i < g_demoEnv.coLi; i++) {
        nQuadE1T1 = 0;
        DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_DUAL_T1E1 #%d)\n", i, nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
      }
    }
    break;

  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }

  if ( (g_demoEnv.nBoardType == OTX_DEVICE_THOR_PCMCIA) || (g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_EXPRESSCARD) || (g_demoEnv.nBoardType == OTX_DEVICE_SLEIPNIR_1_PCI) || (g_demoEnv.nBoardType == OTX_DEVICE_GIMLE_16_PCI_PLUS)) {
    // do nothing, because these boards do not have a OTX_DEVICE_CODEC device
  } else {
    // Get handles to Codec devices #0 - #3
    for (i = 0; i < 4; i++) {
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_CODEC, i, i, g_hEventQueue, &g_hPhysDevs.hCodec[i]); DemoCheckRc(g_hPhysDevs.hThorController0, result);

      // Set correct companded law (u-law or a-law)
      result = OtxBrdCodecSetLaw(g_hPhysDevs.hCodec[i], g_demoEnv.law); DemoCheckRc(g_hPhysDevs.hCodec[i], result);
    }
  }


  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_2_EXPRESSCARD:    
    if(g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM3) {
      // User did not set the module type use default type for this DSP
      g_demoEnv.eProgramModule = OTX_DSP_SPM_OTXSPM2_INTMEM;
    }
    // Get handles to the physcial DSP device
    for (i = 0; i < g_demoEnv.coDsp; i++) {
      char szDevString[32];
      if (i == 0) {
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_DSP, i, i, g_hEventQueue, &g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hVidarAsmController0, result);
        result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hDsp[i], OTX_ATTR_DEVICE_STRING, szDevString, 32); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
        DemoLogf("DSP#%d: %s\n", i, szDevString);
      } else {
        g_hPhysDevs.hDsp[i] = g_hPhysDevs.hDsp[0]; // Just to make a demo program generic for 4 DSPs. This card really only has one DSP.
      }
    } 
    break;
  case OTX_DEVICE_SLEIPNIR_1_PCI:
  case OTX_DEVICE_THOR_PCMCIA:    
    if(g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM3) {
      // User did not set the module type use default type for this DSP
      g_demoEnv.eProgramModule = OTX_DSP_SPM_OTXSPM2;
    }
    // Get handles to the physcial DSP devices #0 and/or #1
    for (i = 0; i < g_demoEnv.coDsp; i++) {
      char szDevString[32];
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_DSP, i, i, g_hEventQueue, &g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hVidarAsmController0, result);
      result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hDsp[i], OTX_ATTR_DEVICE_STRING, szDevString, 32); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
      DemoLogf("DSP#%d: %s\n", i, szDevString);
    } 
    break;
  case OTX_DEVICE_THOR_PCI:
  case OTX_DEVICE_THOR_8_PCI:
  case OTX_DEVICE_THOR_8_PCI_PLUS:
  case OTX_DEVICE_THOR_4_PCI_PLUS:
  case OTX_DEVICE_THOR_2_PCI_PLUS:
  case OTX_DEVICE_THOR_2_PCI_EXPRESS:
  case OTX_DEVICE_GIMLE_16_PCI_PLUS:
    {   
      // Get a handle to the ASM Daughter Board
      OTX_UINT32   eDevType;
      OTX_ITERATOR itrChildTypes;
      
      OtxDrvIteratorBegin(&itrChildTypes);
      while (OTX_SUCCESS(OtxDrvEnumDeviceType(g_hPhysDevs.hThorController0, &itrChildTypes, (OTX_INT32 *)&eDevType))) {  
        OTX_BOOL bOpenAsm = OTX_FALSE;
        switch (eDevType) {
        case OTX_DEVICE_VIDAR_5x_ASM:
          DemoLogf("Opening Vidar-5x4-ASM board\n");
          if(g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM3) {
            // User did not set the module type use default type for this ASM
            g_demoEnv.eProgramModule = OTX_DSP_SPM_OTXSPM2;
          }
          bOpenAsm = OTX_TRUE;
          break;

        case OTX_DEVICE_VIDAR_5X4_ASM_CST:
          DemoLogf("Opening Vidar-5x4-ASM-CST board\n");
          if(g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM3) {
            // User did not set the module type use default type for this ASM
            g_demoEnv.eProgramModule = OTX_DSP_SPM_OTXSPM2;
          }
          bOpenAsm = OTX_TRUE;
          break;
        case OTX_DEVICE_VIDAR_5X4_ASM_EX:
          DemoLogf("Opening Vidar-5x4-ASM-EX board\n");
          if(g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM3) {
            // User did not set the module type use default type for this ASM
            if(g_demoEnv.nBoardType == OTX_DEVICE_GIMLE_16_PCI_PLUS) {
              g_demoEnv.eProgramModule = OTX_DSP_SPM_OTXSPM2EX;
            }
            else {
              g_demoEnv.eProgramModule = OTX_DSP_SPM_OTXSPM2;
            }
          }
          bOpenAsm = OTX_TRUE;
          break;
        case OTX_DEVICE_VIDAR_5X4_ASM_PRO:
          DemoLogf("Opening Vidar-5x4-ASM-PRO board\n");
          if(g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM3) {
            // User did not set the module type use default type for this ASM
            if(g_demoEnv.nBoardType == OTX_DEVICE_GIMLE_16_PCI_PLUS) {
              g_demoEnv.eProgramModule = OTX_DSP_SPM_OTXSPM2EX;
            }
            else {
              g_demoEnv.eProgramModule = OTX_DSP_SPM_OTXSPM2;
            }
          }
          bOpenAsm = OTX_TRUE;
          break;
        case OTX_DEVICE_VIDAR_55x4_ASM:
          DemoLogf("Opening Vidar-55x4-ASM board\n");
          if(g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM3) {
            // User did not set the module type use default type for this ASM
            g_demoEnv.eProgramModule = OTX_DSP_SPM_OTXSPM2;
          }
          bOpenAsm = OTX_TRUE;

          // Verify that the system supports the Vidar-55x4-ASM board
          {
            char szCpldBuildNumbers[256];
            char szBrdBuildNumber[16];
            OTX_INT32 nBoardBuildNo;
          
            if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_PCI) {
            
              //  Verify firmware revision
              result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_CPLD_BUILD, szCpldBuildNumbers, 256); DemoCheckRc(g_hPhysDevs.hThorController0, result);
              if (strstr(szCpldBuildNumbers, "OTXCDC: 0761") || strstr(szCpldBuildNumbers, "OTXCDC: 0762")) {
                DemoLogf("Error: The firmware on the Thor-2-PCI must be upgraded to firmware build 5\n");
                DemoLogf("       (OTXCDC 0764) or later to support the Vidar-55x4-ASM board.\n");
                DemoLogf("       Please download firmware update program from the Odin TeleSystems\n");
                DemoLogf("       web site (www.odints.com)\n");
                DemoLogf("       Also please verify that the HMA-1098-1-2.0 module is plugged into connector BJ3.\n");
                DemoExitQuestion();
              }
            
              // Verify board build number
              result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUILD, szBrdBuildNumber, 16); DemoCheckRc(g_hPhysDevs.hThorController0, result);
              nBoardBuildNo = strtol(strtok(szBrdBuildNumber, "."), NULL, 0);
              DemoLogf("Board Build Number: %d\n", nBoardBuildNo);
              if (nBoardBuildNo < 11) {
                DemoLogf("Error: The Thor-2-PCI board must be upgraded to build 11\n");
                DemoLogf("       or later to support the Vidar-55x4-ASM board.\n");
                DemoLogf("       Please contact Odin TeleSystems Support Team for upgrade details.\n");
                DemoExitQuestion();
              }
            
            } else if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI) {
              //  Verify firmware revision
              result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_CPLD_BUILD, szCpldBuildNumbers, 256); DemoCheckRc(g_hPhysDevs.hThorController0, result);
              if (strstr(szCpldBuildNumbers, "OTXCDC: 0771")) {
                DemoLogf("Error: The firmware on the Thor-8-PCI must be upgraded to firmware build 3\n");
                DemoLogf("       (OTXCDC 0772) or later to support the Vidar-55x4-ASM board.\n");
                DemoLogf("       Please download firmware update program from the Odin TeleSystems\n");
                DemoLogf("       web site (www.odints.com)\n");
                DemoLogf("       Also please verify that the HMA-1098-1-2.0 module is plugged into connector BJ3.\n");
                DemoExitQuestion();
              }
            
              // Verify board build number
              result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUILD, szBrdBuildNumber, 16); DemoCheckRc(g_hPhysDevs.hThorController0, result);
              nBoardBuildNo = strtol(szBrdBuildNumber, NULL, 0);
              DemoLogf("Board Build Number: %d\n", nBoardBuildNo);
              if (nBoardBuildNo < 5) {
                DemoLogf("Error: The Thor-8-PCI board must be upgraded to build 5\n");
                DemoLogf("       or later to support the Vidar-55x4-ASM board.\n");
                DemoLogf("       Please contact Odin TeleSystems Support Team for upgrade details.\n");
                DemoExitQuestion();
              }            
            }          
          }

          break;

        // ASM boards not supported
        case OTX_DEVICE_VIDAR_5X4_ASM_V90:

          if(eDevType == OTX_DEVICE_VIDAR_5X4_ASM_V90) {
            DemoLogf("Vidar-5x4-ASM-V90 board\n");
          }

          DemoLogf("ASM board not supported, %d\n", eDevType);
          DemoExitQuestion();
          // Did not exit. Need to set an ASM load module.
          if(g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM3) {
            // User did not set the module type use default type for this ASM
            g_demoEnv.eProgramModule = OTX_DSP_SPM_OTXSPM2EX;
          }
          bOpenAsm = OTX_TRUE;
          break;

        default:
          break;
        }
        if (bOpenAsm) {
          // Get a handle to to the ASM Daughter Board
          result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, (OtxPhysicalDeviceTypeE) eDevType, 0, 0, g_hEventQueue, &g_hPhysDevs.hVidarAsmController0); DemoCheckRc(g_hPhysDevs.hThorController0, result);
        }
      
      } //while
      
      for (i = 0; i < g_demoEnv.coDsp; i++) {
        char szDevString[32];
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hVidarAsmController0, OTX_DEVICE_DSP, i, i, g_hEventQueue, &g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hVidarAsmController0, result);
        result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hDsp[i], OTX_ATTR_DEVICE_STRING, szDevString, 32); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
        DemoLogf("DSP#%d: %s\n", i, szDevString);
      } 
    }
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }

  // Get DSP Highway Speed
  for (i = 0; i < g_demoEnv.coDsp; i++) {
    OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[i], OTX_ATTR_DSP_HIGHWAY_SPEED, &g_nDspHighwaySpeed); 
    DemoLogf("DSP#%d highway speed is %d kbps\n", i, g_nDspHighwaySpeed);

    if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_EXPRESSCARD) break; // There is only one DSP on this card. No need to load the program more than once.
  }


  // Load DSP Program
  if (g_demoEnv.pszDspOutFile) {
    // Load from file
    for (nDsp = 0; nDsp < g_demoEnv.coDsp; nDsp++) {
      DemoLogf("Loading program file %s to DSP#%d \n", g_demoEnv.pszDspOutFile,  nDsp);
      result = OtxDspRunProgram(g_hPhysDevs.hDsp[nDsp], g_demoEnv.pszDspOutFile, "_c_int00"); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);

      if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_EXPRESSCARD) break; // There is only one DSP on this card. No need to load the program more than once.
    }

  } else {
    
    // Load DSP Program from OtxDrv.lib
    for (nDsp = 0; nDsp < g_demoEnv.coDsp; nDsp++) {
      switch (g_demoEnv.eProgramModule) {
      case OTX_DSP_SPM_OTXSPM1:
        DemoLogf("DSP#%d: Loading DSP program OTX_DSP_SPM_OTXSPM1\n", nDsp);
        break;
      case OTX_DSP_SPM_OTXSPM2:
        DemoLogf("DSP#%d: Loading DSP program OTX_DSP_SPM_OTXSPM2\n", nDsp);
        break;
      case OTX_DSP_SPM_OTXSPM2EX:
        DemoLogf("DSP#%d: Loading DSP program OTX_DSP_SPM_OTXSPM2EX\n", nDsp);
        break;
      case OTX_DSP_SPM_OTXSPM2_EXTMEM:
        DemoLogf("DSP#%d: Loading DSP program OTX_DSP_SPM_OTXSPM2_EXTMEM\n", nDsp);
        break;
      case OTX_DSP_SPM_OTXSPM2_INTMEM:
        DemoLogf("DSP#%d: Loading DSP program OTX_DSP_SPM_OTXSPM2_INTMEM\n", nDsp);
        break;
      case OTX_DSP_SPM_OTXSPM2EX_INTMEM:
        DemoLogf("DSP#%d: Loading DSP program OTX_DSP_SPM_OTXSPM2EX_INTMEM\n", nDsp);
        break;
      case OTX_DSP_SPM_OTXHDLC:
        DemoLogf("Loading DSP program OTX_DSP_SPM_OTXHDLC is not implemented for this demo program.\n");
        OtxDrvReleaseLib();
        OsaConsoleExit();
        exit(1);
        break;
      case OTX_DSP_SPM_OTXDTTX:
        DemoLogf("DSP#%d: Loading DSP program OTX_DSP_SPM_OTXDTTX\n", nDsp);
        break;
      default:
        DemoLogf("Unknown DSP program module (%d). Exiting.\n", g_demoEnv.eProgramModule);
        OtxDrvReleaseLib();
        OsaConsoleExit();
        exit(1);
      }
      
      result = OtxDspRunSpm(g_hPhysDevs.hDsp[nDsp], g_demoEnv.eProgramModule); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);

      if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_EXPRESSCARD) break; // There is only one DSP on this card. No need to load the program more than once.
    }
  }

#if 0 // Create an OTX DSP Image from the DSP Out-file, then load the image (faster)
  {
    OTX_UINT16 *psqProgImage;
    OTX_UINT16 coProgImage;
    OTX_INT32  nEntryAddr;
    const OTX_UINT16 DEMO_DSP_IMAGE_BUF_SIZE = 60000;
    
    psqProgImage = (OTX_UINT16 *)malloc(DEMO_DSP_IMAGE_BUF_SIZE*sizeof(OTX_UINT16));
    if (!psqProgImage) {
      DemoLogf("Could not allocate buffer for DSP program image. Exiting.\n");
      OtxDrvReleaseLib();
      OsaConsoleExit();
			exit(1);
    }
    
    result = OtxDspCreateProgramImage("OtxSpm2.out", "_c_int00", DEMO_DSP_IMAGE_BUF_SIZE, psqProgImage, &coProgImage, &nEntryAddr); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    for (nDsp = 0; nDsp < g_demoEnv.coDsp; nDsp++) {
      DemoLogf("Loading image for OtxSpm2.out to DSP#%d (coProgImage=%d, nEntryAddr=%d) \n", nDsp, coProgImage, nEntryAddr);
      result = OtxDspRunImage(g_hPhysDevs.hDsp[nDsp], psqProgImage, coProgImage, nEntryAddr); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
    }
    
    free(psqProgImage);
  }
#endif

  if (g_demoEnv.bUseDspDebugger) {
    int c;
    DemoLogf("************************************************************\n\n");
    DemoLogf("Now please load and run DSP program to all of the DSPs\n");
    DemoLogf("in your DSP Debugger.\n\n");
    DemoLogf("After all DSPs are running:\n");
    DemoLogf("  Press 'r' to resynchronize the communication between the DSPs and the host\n");
    DemoLogf("  Press 's' to skip the resynchronization step\n");
    c = OsaConsoleGetCh();
    DemoLogf("\n");

    if (c == 'r' || c == 'R') {
      // At this point we need to tell the host to restart the programs 
      // (to issue the initial IO commands to the DSP again)
      for (i = 0; i < g_demoEnv.coDsp; i++) {
        DemoLogf("Restarting program on DSP#%d\n", i);
        result = OtxDspRestartProgram(g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
      }
      // At this point the DSP should be in the same state as they were right after OtxDspRunProgram() was called
    }
  }


}


/*******************************************************************************

  @func
    Configure E1/T1 Devices

  @comm The clock source of the board(s) must set by a call 
  to <f OtxBrdSetClocks> prior to calling this function.

*******************************************************************************/
void DemoConfigureE1T1Devices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  // Set default configuration data
  DemoInitT1E1ConfigData( &(g_T1E1CfgData.e1CfgData), &(g_T1E1CfgData.t1CfgData));
  if(g_demoEnv.nBoardType == OTX_DEVICE_GIMLE_16_PCI_PLUS) {
    for (i = 0; i < g_demoEnv.coLi; i++) {
      if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1){ // T1
        // Configure T1 Transceivers
        result = OtxT1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
        // Set them in T1 mode
        result =  OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], OTX_T1E1_MODE_T1); DemoCheckRc(g_hPhysDevs.hLi[i], result);
      }
      else{  // E1
        // Configure E1 Transceivers
        result = OtxE1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.e1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
        // Set them in E1 mode
        result =  OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], OTX_T1E1_MODE_E1); DemoCheckRc(g_hPhysDevs.hLi[i], result);
      }

      // Tri-state the transmit pins of the Line interface (so it does not disturb the receive pins in High-Z mode)
      result = OtxT1E1XmitPinControl(g_hPhysDevs.hLi[i], OTX_FALSE); DemoCheckRc(g_hPhysDevs.hLi[i], result);

      // Set line termination mode
      {

        DemoLogf("Li[%d]: Setting Line Termination mode to ", i);
        switch (g_demoEnv.eTermMode) {
        case OTX_T1E1_TERM_120_OHM:
          DemoLogf("OTX_T1E1_TERM_120_OHM\n");
          break;
        case OTX_T1E1_TERM_100_OHM:
          DemoLogf("OTX_T1E1_TERM_100_OHM\n");
          break;
        case OTX_T1E1_TERM_75_OHM:
          DemoLogf("OTX_T1E1_TERM_75_OHM\n");
          break;
        case OTX_T1E1_TERM_HIGHZ:
          DemoLogf("OTX_T1E1_TERM_HIGHZ\n");
          break;
        default:
          DemoLogf("Invalid Termination mode.\n");
          break;
        }

        result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[i], g_demoEnv.eTermMode); DemoCheckRc(g_hPhysDevs.hLi[i], result);
      }

      // Set receive threshold
      DemoLogf("Li[%d]: Setting RX Input Threshold to %d\n", i, g_demoEnv.nRxThreshold);
      result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hLi[i], OTX_ATTR_T1E1_RX_INPUT_THRESHOLD, g_demoEnv.nRxThreshold); DemoCheckRc(g_hPhysDevs.hLi[i], result);
      { // Read back the threshold
        OTX_UINT32 nRxThreshold;
        result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hLi[i], OTX_ATTR_T1E1_RX_INPUT_THRESHOLD, &nRxThreshold); DemoCheckRc(g_hPhysDevs.hLi[i], result);
        DemoLogf("Li[%d]: RX Input Threshold = %d\n", i, nRxThreshold);
      }
    
      // Set Monitor (-20dB) mode
      DemoLogf("Li[%d]: Setting Monitor mode to %s\n", i, g_demoEnv.bMonitorMode ? "ON" : "OFF");
      result = OtxT1E1SetMonitorMode(g_hPhysDevs.hLi[i], g_demoEnv.bMonitorMode); DemoCheckRc(g_hPhysDevs.hLi[i], result);
    
      // Turning front end Amplifier ON/OFF      
      DemoLogf("Li[%d]: Setting Front End Amplifier to %s\n", i, g_demoEnv.bAmplifier ? "ON" : "OFF");
      result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hLi[i], OTX_ATTR_T1E1_RX_AMPLIFIER, g_demoEnv.bAmplifier); DemoCheckRc(g_hPhysDevs.hLi[i], result);
    }

  } else {
    // Configure Transceivers
    for (i = 0; i < g_demoEnv.coLi; i++) {

      // Set termination for the T1/E1 Front End
      result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[i], OTX_T1E1_TERM_100_OHM); DemoCheckRc(g_hPhysDevs.hLi[i], result);

      if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
        // T1 mode
        result = OtxT1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
      } else {
        // E1 mode
        result = OtxE1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.e1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
      }
      result = OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], g_demoEnv.eT1E1Mode); DemoCheckRc(g_hPhysDevs.hLi[i], result);

      if (g_demoEnv.bLiLoop) {
        // Tell the E1T1 device to loop its internal highway back to the internal highway
        result = OtxT1E1SetLineLoop(g_hPhysDevs.hLi[i], OTX_T1E1_LOCAL_LOOP); DemoCheckRc(g_hPhysDevs.hLi[i], result);
      }

    }
  }

}



/*******************************************************************************

  @func
    Enable Configured Physical Devices

*******************************************************************************/
void DemoEnablePhysicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;
  OTX_INT32  nQuadE1T1; // Only used for Thor-4-PCI-Plus, Thor-8-PCI-(Plus)

  result = OtxDrvEnable(g_hPhysDevs.hThorController0,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hThorController0, result);
  if (g_hPhysDevs.hVidarAsmController0) {
    result = OtxDrvEnable(g_hPhysDevs.hVidarAsmController0, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hVidarAsmController0, result);
  }

  // If board type is Thor-8-PCI-(Plus) then enable Quad E1T1 device
  if ((g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI) || (g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI_PLUS)) {
    OTX_INT32 nQuadE1T1; // Only used for Thor-8-PCI
      
    // Enable Quad E1T1 devices
    for (nQuadE1T1 = 0; nQuadE1T1 < 2; nQuadE1T1++) {
      DemoLogf("Enable OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
      result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
    }
  }
    
  // If board type is Thor-4-PCI-Plus then enable Quad E1T1 device
  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_PCI_PLUS) {
      
    // Enable Quad E1T1 devices
    for (nQuadE1T1 = 0; nQuadE1T1 < 1; nQuadE1T1++) {
      DemoLogf("Enable OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
      result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
    }
  }

  // If board type is Gimle-16-PCI-Plus then enable Quad E1T1 device
  if (g_demoEnv.nBoardType == OTX_DEVICE_GIMLE_16_PCI_PLUS) {

    // Enable Quad E1T1 devices
    for (nQuadE1T1 = 0; nQuadE1T1 < 4; nQuadE1T1++) {
      DemoLogf("Enable OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
      result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
    }
  }
  
  // If board type is Thor-2-ExpressCard then enable Dual E1T1 device
  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_EXPRESSCARD) {
      
    // Enable Quad E1T1 devices
    nQuadE1T1 = 0;
    DemoLogf("Enable OTX_DEVICE_DUAL_T1E1 #%d device\n", nQuadE1T1);
    result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
  }


  for (i = 0; i < g_demoEnv.coLi; i++) {
    result = OtxDrvEnable(g_hPhysDevs.hLi[i],            OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hLi[i], result);
  }

  for (i = 0; i < g_demoEnv.coDsp; i++) {
    result = OtxDrvEnable(g_hPhysDevs.hDsp[i],            OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
  }

}



/*******************************************************************************

  @func
    Create the needed Logical Devices

  @comm
    The logical devices are differenciated from the phyiscal devices by the fact
    the number of logical devices is not set. More of them can be requested
    by the user and allocated by the system if needed.

*******************************************************************************/
void DemoCreateLogicalDevices(
  void
)
{
  OTX_RESULT result;
  OTX_INT32  i;

  for (i = 0; i < g_demoEnv.coLi; i++) {
    // Create Hdlc Senders and Receivers logical devices hosted in the Dsp devices
    result = OtxDrvCreateLogicalDevice( g_hPhysDevs.hDsp[i], OTX_LDEVICE_HDLC_SENDER, g_hEventQueue, i, &g_hLogicDevs.hHdlcSender[i]); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
    result = OtxDrvCreateLogicalDevice( g_hPhysDevs.hDsp[i], OTX_LDEVICE_HDLC_RECEIVER, g_hEventQueue, i, &g_hLogicDevs.hHdlcReceiver[i]); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
  }


  for (i = 0; i < g_demoEnv.coDsp; i++) {
    if (g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM1 || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2 || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2_INTMEM || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2EX_INTMEM || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2EX || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2_EXTMEM) {
      // Create Dtmf Detector Logical Devices
#ifdef DEMO_DUAL_TONE_DETECTOR
      result = OtxDrvCreateLogicalDevice( g_hPhysDevs.hDsp[i], OTX_LDEVICE_TONE_DUAL_DETECTOR, g_hEventQueue, i, &g_hLogicDevs.hDtmfDetector[i]); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
#else
      result = OtxDrvCreateLogicalDevice( g_hPhysDevs.hDsp[i], OTX_LDEVICE_TONE_DTMF_DETECTOR, g_hEventQueue, i, &g_hLogicDevs.hDtmfDetector[i]); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
#endif

      // Create Sinewave Sender Logical Devices
      result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[i], OTX_LDEVICE_TONE_SINEWAVE_GENERATOR, g_hEventQueue, i, &g_hLogicDevs.hSineWaveSender[i]); DemoCheckRc(g_hPhysDevs.hDsp[i], result);

    }

    // Create DTMF Dialers Logical Devices
    result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[i], OTX_LDEVICE_TONE_DTMF_DIALER, g_hEventQueue, i, &g_hLogicDevs.hDtmfDialler[i]); DemoCheckRc(g_hPhysDevs.hDsp[i], result);

    if (g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2 || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2EX || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2_INTMEM || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2EX_INTMEM || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2_EXTMEM) {
      // Create Sinewave Receiver Logical Devices (use a DTMF detector as a device to test the Sinewave Generator)
      result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[i], OTX_LDEVICE_TONE_MF_DETECTOR, g_hEventQueue, i, &g_hLogicDevs.hSineWaveReceiver[i]); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
    } else if (g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM1) {
      // Create Sinewave Receiver Logical Devices (use a DTMF detector as a device to test the Sinewave Generator)
      result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[i], OTX_LDEVICE_TONE_DTMF_DETECTOR, g_hEventQueue, i, &g_hLogicDevs.hSineWaveReceiver[i]); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
    }

    if (g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2 || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2EX || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2_INTMEM || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2EX_INTMEM || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2_EXTMEM) {
      // Create MF Dialers Logical Devices
      result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[i], OTX_LDEVICE_TONE_MF_DIALER, g_hEventQueue, i, &g_hLogicDevs.hMfDialler[i]); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
    }

  }

}

/*******************************************************************************

  @func
    Close Logical Devices

*******************************************************************************/
void DemoCloseLogicalDevices(
  void
)
{
  OTX_RESULT result;
  OTX_INT32  i;

  for (i = 0; i < g_demoEnv.coLi; i++) {
    // Close Hdlc Senders and Receivers logical devices hosted in the physical Line Interface devices
    result = OtxDrvClose(g_hLogicDevs.hHdlcSender[i]); DemoCheckRc(g_hLogicDevs.hHdlcSender[i], result);
    result = OtxDrvClose(g_hLogicDevs.hHdlcReceiver[i]); DemoCheckRc(g_hLogicDevs.hHdlcReceiver[i], result);
  }


  for (i = 0; i < g_demoEnv.coDsp; i++) {
    if (g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM1 || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2 || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2EX || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2_INTMEM || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2EX_INTMEM || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2_EXTMEM) {

      // Close Dtmf Detector Logical Devices
      result = OtxDrvClose(g_hLogicDevs.hDtmfDetector[i]); DemoCheckRc(g_hLogicDevs.hDtmfDetector[i], result);

      // Close Sinewave Sender Logical Devices
      result = OtxDrvClose(g_hLogicDevs.hSineWaveSender[i]); DemoCheckRc(g_hLogicDevs.hSineWaveSender[i], result);

      // Close Sinewave Receiver Logical Devices
      result = OtxDrvClose(g_hLogicDevs.hSineWaveReceiver[i]); DemoCheckRc(g_hLogicDevs.hSineWaveReceiver[i], result);

    }

    // Close DTMF Dialers Logical Devices
    result = OtxDrvClose(g_hLogicDevs.hDtmfDialler[i]); DemoCheckRc(g_hLogicDevs.hDtmfDialler[i], result);

    if (g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2 || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2EX || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2_INTMEM || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2EX_INTMEM || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2_EXTMEM) {
      // Close MF Dialers Logical Devices
      result = OtxDrvClose(g_hLogicDevs.hMfDialler[i]); DemoCheckRc(g_hLogicDevs.hMfDialler[i], result);
    }

  }

}


/*******************************************************************************

  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset of a
    Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a characterized
    bu its capacity (in 8 kbit/s increments). The user can create a pipe with
    the driver function <f OtxDrvCreatePipe>. A pipe is connected to physical or
    logical devices and mapped to specific time-slots with the driver function
    <f OtxDrvConnectPipe>.

*******************************************************************************/
void DemoConnectDevices(
  void
)
{
  OTX_INT32 nLi;
  OTX_INT32 nDsp;
  OTX_INT32 i,nDspNo;
  OTX_RESULT result;
  OTX_UINT8 sqLiTimeSlotMask[32];
  OTX_UINT8 sqDspTimeSlotMask[128];
  OTX_INT32 nTsLi;
  OTX_INT32 nTsDsp;
  OTX_UINT32 coTsPerDspHwy;
  OTX_UINT32 nDspHighwaySpeed;

  if (g_demoEnv.nBoardType == OTX_DEVICE_SLEIPNIR_1_PCI) {
    nDspNo = g_demoEnv.coDsp;
  } else {
    nDspNo = g_demoEnv.coDsp/2;
  }

  // Connect HDLC pipes
  for (i = 0; i < g_demoEnv.coLi; i++) {

    nLi = i;
    nDsp = i;

    // Create an incoming and outgoing Pipe representing the external T1/E1
    result = OtxDrvCreatePipe(64, &g_hPipes.hLiIn[i]); DemoCheckRc(g_hPipes.hLiIn[i], result);
    result = OtxDrvCreatePipe(64, &g_hPipes.hLiOut[i]); DemoCheckRc(g_hPipes.hLiOut[i], result);

    // Get DSP Highway Speed
    OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_HIGHWAY_SPEED, &nDspHighwaySpeed); 
    DemoLogf("DSP#%d highway speed is %d kbps\n", nDsp, nDspHighwaySpeed);
    coTsPerDspHwy = nDspHighwaySpeed / 64;

    if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
      // T1 mode
      nTsLi = 24; 
    } else {
      // E1 mode
      nTsLi = 16; 
    }
    if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_EXPRESSCARD) {
      // The Li0 timeslots on this card are the TS0, TS4, TS8, ... , TS124 timeslots of the DSP highway.
      // The Li1 timeslots on this card are the TS2, TS6, TS10, ... , TS126 timeslots of the DSP highway.    
      nTsDsp = nTsLi * 4 + nLi * 2 ;  // See comment above regarding the timeslot mapping for the DSP highway on this board
    } else {
      nTsDsp = 127;
      //nTsDsp = nTsLi;
    }
    memset(sqLiTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
    sqLiTimeSlotMask[nTsLi] = 0xFF;
    memset(sqDspTimeSlotMask, 0, coTsPerDspHwy * sizeof(OTX_UINT8));
    sqDspTimeSlotMask[nTsDsp] = 0xFF;

    DemoLogf("Connecting HDLC device[%d]: nLi=%d, nTsLi=%.2d, nDsp=%d, nTsDsp=%.3d\n", i, nLi, nTsLi, nDsp, nTsDsp);

    result = OtxDrvConnectPipe(g_hPipes.hLiOut[i], OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[i], -1, sqLiTimeSlotMask, 32); DemoCheckRc(g_hPipes.hLiOut[i], result);
    //if ( (g_demoEnv.nBoardType == OTX_DEVICE_THOR_PCMCIA) || (g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_EXPRESSCARD) || (g_demoEnv.nBoardType == OTX_DEVICE_SLEIPNIR_1_PCI) ) {
      // Connection for boards without a time-space switch. We specify the timeslot for the dsp highway connection.
      result = OtxDrvConnectPipe(g_hPipes.hLiOut[i], OTX_PIPE_INPUT, g_hPhysDevs.hDsp[i], -1, sqDspTimeSlotMask, coTsPerDspHwy); DemoCheckRc(g_hPipes.hLiOut[i], result);
#if 0
    } else {
      // Connection for boards with a time-space switch. We let the driver pick the timeslot for the dsp highway connection (no timeslotmask)
      result = OtxDrvConnectPipe(g_hPipes.hLiOut[i], OTX_PIPE_INPUT, g_hPhysDevs.hDsp[i], -1, NULL, 0); DemoCheckRc(g_hPipes.hLiOut[i], result);
    }
#endif

    result = OtxDrvConnectPipe(g_hPipes.hLiIn[i], OTX_PIPE_INPUT,  g_hPhysDevs.hLi[i], -1, sqLiTimeSlotMask, 32); DemoCheckRc(g_hPipes.hLiIn[i], result);
    //if ( (g_demoEnv.nBoardType == OTX_DEVICE_THOR_PCMCIA) || (g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_EXPRESSCARD) || (g_demoEnv.nBoardType == OTX_DEVICE_SLEIPNIR_1_PCI) ) {
      // Connection for boards without a time-space switch. We specify the timeslot for the dsp highway connection.
      result = OtxDrvConnectPipe(g_hPipes.hLiIn[i], OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[i], -1, sqDspTimeSlotMask, coTsPerDspHwy); DemoCheckRc(g_hPipes.hLiIn[i], result);
#if 0
    } else {
      // Connection for boards with a time-space switch. We let the driver pick the timeslot for the dsp highway connection (no timeslotmask)
      result = OtxDrvConnectPipe(g_hPipes.hLiIn[i], OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[i], -1, NULL, 0); DemoCheckRc(g_hPipes.hLiIn[i], result);
    }
#endif

    // Connect HDLC Sender and Receiver devices to the pipes
    result = OtxDrvConnectLogicalDevice(g_hPipes.hLiIn[i], -1, g_hLogicDevs.hHdlcReceiver[i], -1); DemoCheckRc(g_hPipes.hLiIn[i], result);
    result = OtxDrvConnectLogicalDevice(g_hLogicDevs.hHdlcSender[i], -1, g_hPipes.hLiOut[i], -1); DemoCheckRc(g_hLogicDevs.hHdlcSender[i], result);

  }

  // Create pipe for tone devices
  for (i = 0; i < g_demoEnv.coDsp; i++) {
    result = OtxDrvCreatePipe(64, &g_hPipes.hLiDspDtmfDetector[i]); DemoCheckRc(g_hPipes.hLiDspDtmfDetector[i], result);
    result = OtxDrvCreatePipe(64, &g_hPipes.hDspLiDtmfDialer[i]); DemoCheckRc(g_hPipes.hDspLiDtmfDialer[i], result);

    result = OtxDrvCreatePipe(64, &g_hPipes.hLiDspSineWaveReceiver[i]); DemoCheckRc(g_hPipes.hLiDspSineWaveReceiver[i], result);
    result = OtxDrvCreatePipe(64, &g_hPipes.hDspLiSineWaveSender[i]); DemoCheckRc(g_hPipes.hDspLiSineWaveSender[i], result);

  }                  

  // Connect pipes for the DTMF Dialer and Detector

  for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {

    // Create two pipes per Li. If four DSPs are available (e.g. on the Vidar-ASM boards) we will distribute
    // the connections in the following way:
    // Li0 Ts1 <=> DSP0
    // Li0 Ts2 <=> DSP1
    // Li1 Ts1 <=> DSP2
    // Li1 Ts2 <=> DSP3
    // If four DSPs are not available (e.g. Thor-2-PCMCIA has 2 DSPs, Thor-2-ExpressCard and Sleipnir-1-PCI has 1 DSP)
    // we will distribute the pipe connections over the available DSPs.
    for (i = 0; i < nDspNo; i++) {
      
      nDsp = i+nLi*(g_demoEnv.coDsp/2);
      
      // Get DSP Highway Speed
      OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_HIGHWAY_SPEED, &nDspHighwaySpeed); 
      DemoLogf("DSP#%d highway speed is %d kbps\n", nDsp, nDspHighwaySpeed);
      coTsPerDspHwy = nDspHighwaySpeed / 64;
      /*
      {
        OTX_UINT32 nAttrVal;
        OtxDrvGetAttributeValueUint32Ex(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_HWDEVICE_HWY_NO_OF_TS, 0, &nAttrVal); 
        DemoLogf("DSP#%d highway[0] has %d timeslots\n", nDsp, nAttrVal);
        OtxDrvGetAttributeValueUint32Ex(g_hPhysDevs.hLi[nLi], OTX_ATTR_HWDEVICE_HWY_NO_OF_TS, 0, &nAttrVal); 
        DemoLogf("Li#%d highway[0] has %d timeslots\n", nLi, nAttrVal);
      }
      */

      // Connect Pipes
      nTsLi  = 1+i;
      if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_EXPRESSCARD) {
        // The Li0 timeslots on this card are the TS0, TS4, TS8, ... , TS124 timeslots of the DSP highway.
        // The Li1 timeslots on this card are the TS2, TS6, TS10, ... , TS126 timeslots of the DSP highway.    
        nTsDsp = nTsLi * 4 + nLi * 2 ;  // See comment above regarding the timeslot mapping for the DSP highway on this board
      } else {
        nTsDsp = nTsLi;
      }
      memset(sqLiTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
      sqLiTimeSlotMask[nTsLi] = 0xFF;
      memset(sqDspTimeSlotMask, 0, coTsPerDspHwy * sizeof(OTX_UINT8));
      sqDspTimeSlotMask[nTsDsp] = 0xFF;
      
      result = OtxDrvConnectPipe(g_hPipes.hLiDspDtmfDetector[nDsp], OTX_PIPE_INPUT,  g_hPhysDevs.hLi[nLi],  -1, sqLiTimeSlotMask, 32); DemoCheckRc(g_hPipes.hLiDspDtmfDetector[nDsp], result);
      result = OtxDrvConnectPipe(g_hPipes.hDspLiDtmfDialer[nDsp], OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[nLi],  -1, sqLiTimeSlotMask, 32); DemoCheckRc(g_hPipes.hDspLiDtmfDialer[nDsp], result);
      
      if ( (g_demoEnv.nBoardType == OTX_DEVICE_THOR_PCMCIA) || (g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_EXPRESSCARD) || (g_demoEnv.nBoardType == OTX_DEVICE_SLEIPNIR_1_PCI) ) {
        // Connection for boards without a time-space switch. We specify the timeslot for the dsp highway connection.
        result = OtxDrvConnectPipe(g_hPipes.hLiDspDtmfDetector[nDsp], OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[nDsp], -1, sqDspTimeSlotMask, coTsPerDspHwy); DemoCheckRc(g_hPipes.hLiDspDtmfDetector[nDsp], result);
        result = OtxDrvConnectPipe(g_hPipes.hDspLiDtmfDialer[nDsp], OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[nDsp], -1, sqDspTimeSlotMask, coTsPerDspHwy); DemoCheckRc(g_hPipes.hDspLiDtmfDialer[nDsp], result);
      } else {
        // Connection for boards with a time-space switch. We let the driver pick the timeslot for the dsp highway connection (no timeslotmask)
        result = OtxDrvConnectPipe(g_hPipes.hLiDspDtmfDetector[nDsp], OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[nDsp], -1, NULL, 0); DemoCheckRc(g_hPipes.hLiDspDtmfDetector[nDsp], result);
        result = OtxDrvConnectPipe(g_hPipes.hDspLiDtmfDialer[nDsp], OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[nDsp], -1, NULL, 0); DemoCheckRc(g_hPipes.hDspLiDtmfDialer[nDsp], result);
      }
    }
  }

  // Connect pipes for the SineWave Sender and Receiver
  for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
    
    // Connect pipes for the SineWave Sender and Receiver
    // Create two pipes per Li. One pipe to each DSP. If four DSPs are available (e.g. on the Vidar-ASM boards) we will distribute
    // the connections in the following way
    // Li0 Ts5 <=> DSP0
    // Li0 Ts6 <=> DSP1
    // Li1 Ts5 <=> DSP2
    // Li1 Ts6 <=> DSP3
    // If four DSPs are not available (e.g. Thor-2-PCMCIA has 2 DSPs, Thor-2-ExpressCard and Sleipnir-1-PCI has 1 DSP)
    // we will distribute the pipe connections over the available DSPs.
    
    for (i = 0; i < nDspNo; i++) {
      nDsp = i+nLi*(g_demoEnv.coDsp/2);
      
      // Get DSP Highway Speed
      OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_HIGHWAY_SPEED, &nDspHighwaySpeed); 
      DemoLogf("DSP#%d highway speed is %d kbps\n", nDsp, nDspHighwaySpeed);
      coTsPerDspHwy = nDspHighwaySpeed / 64;
      
      // Connect Pipes
      nTsLi = 5+i;
      if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_EXPRESSCARD) {
        // The Li0 timeslots on this card are the TS0, TS4, TS8, ... , TS124 timeslots of the DSP highway.
        // The Li1 timeslots on this card are the TS2, TS6, TS10, ... , TS126 timeslots of the DSP highway.    
        nTsDsp = nTsLi * 4 + nLi * 2 ;  // See comment above regarding the timeslot mapping for the DSP highway on this board
      } else {
        nTsDsp = nTsLi;
      }
      memset(sqLiTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
      sqLiTimeSlotMask[nTsLi] = 0xFF;
      memset(sqDspTimeSlotMask, 0, coTsPerDspHwy * sizeof(OTX_UINT8));
      sqDspTimeSlotMask[nTsDsp] = 0xFF;
      
      result = OtxDrvConnectPipe(g_hPipes.hLiDspSineWaveReceiver[nDsp], OTX_PIPE_INPUT,  g_hPhysDevs.hLi[nLi],  -1, sqLiTimeSlotMask, 32); DemoCheckRc(g_hPipes.hLiDspSineWaveReceiver[nDsp], result);
      result = OtxDrvConnectPipe(g_hPipes.hDspLiSineWaveSender[nDsp], OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[nLi],  -1, sqLiTimeSlotMask, 32); DemoCheckRc(g_hPipes.hDspLiSineWaveSender[nDsp], result);
      
      if ( (g_demoEnv.nBoardType == OTX_DEVICE_THOR_PCMCIA) || (g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_EXPRESSCARD) || (g_demoEnv.nBoardType == OTX_DEVICE_SLEIPNIR_1_PCI) ) {
        // Connection for boards without a time-space switch. We specify the timeslot for the dsp highway connection.
        result = OtxDrvConnectPipe(g_hPipes.hLiDspSineWaveReceiver[nDsp], OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[nDsp], -1, sqDspTimeSlotMask, coTsPerDspHwy); DemoCheckRc(g_hPipes.hLiDspSineWaveReceiver[nDsp], result);
        result = OtxDrvConnectPipe(g_hPipes.hDspLiSineWaveSender[nDsp], OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[nDsp], -1, sqDspTimeSlotMask, coTsPerDspHwy); DemoCheckRc(g_hPipes.hDspLiSineWaveSender[nDsp], result);
      } else {
        // Connection for boards with a time-space switch. We let the driver pick the timeslot for the dsp highway connection (no timeslotmask)
        result = OtxDrvConnectPipe(g_hPipes.hLiDspSineWaveReceiver[nDsp], OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[nDsp], -1, NULL, 0); DemoCheckRc(g_hPipes.hLiDspSineWaveReceiver[nDsp], result);
        result = OtxDrvConnectPipe(g_hPipes.hDspLiSineWaveSender[nDsp], OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[nDsp], -1, NULL, 0); DemoCheckRc(g_hPipes.hDspLiSineWaveSender[nDsp], result);
      }
    }
    
    
  }


  for (i = 0; i < g_demoEnv.coDsp; i++) {

    // CONNECT LOGICAL DEVICES WITHIN THE DSP
    // --------------------------------------

    // Create the following Connections in receive direction:
    //  Li  ------> DtmfDetector
    if (g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM1 || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2 || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2EX || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2_INTMEM || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2EX_INTMEM || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2_EXTMEM) {
      result = OtxDrvConnectLogicalDevice(g_hPipes.hLiDspDtmfDetector[i],     -1, g_hLogicDevs.hDtmfDetector[i],       -1); DemoCheckRc(g_hPipes.hLiDspDtmfDetector[i], result);
      result = OtxDrvConnectLogicalDevice(g_hPipes.hLiDspSineWaveReceiver[i], -1, g_hLogicDevs.hSineWaveReceiver[i],   -1); DemoCheckRc(g_hPipes.hLiDspSineWaveReceiver[i], result);
    }

    // Make the following connections in transmit direction:
    //              DtmfDialler   -----> Li
    result = OtxDrvConnectLogicalDevice(g_hLogicDevs.hDtmfDialler[i], -1, g_hPipes.hDspLiDtmfDialer[i], -1); DemoCheckRc(g_hLogicDevs.hDtmfDialler[i], result);

    if (g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM1 || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2 || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2EX || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2_INTMEM || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2EX_INTMEM || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2_EXTMEM) {
      result = OtxDrvConnectLogicalDevice(g_hLogicDevs.hSineWaveSender[i], -1, g_hPipes.hDspLiSineWaveSender[i], -1); DemoCheckRc(g_hLogicDevs.hSineWaveSender[i], result);
    }


    if (g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2 || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2EX || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2_INTMEM || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2EX_INTMEM || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2_EXTMEM) {
      result = OtxDrvConnectLogicalDevice(g_hLogicDevs.hMfDialler[i],      -1, g_hPipes.hDspLiSineWaveSender[i], -1); DemoCheckRc(g_hLogicDevs.hMfDialler[i], result);
    }

  }

  // Connect Codecs (speakers)
  if ( (g_demoEnv.nBoardType == OTX_DEVICE_THOR_PCMCIA) || (g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_EXPRESSCARD) || (g_demoEnv.nBoardType == OTX_DEVICE_SLEIPNIR_1_PCI) ) {
    // do nothing, because they don't have OTX_DEVICE_CODEC on these boards
  } else {
    for (nDsp = 0; nDsp < g_demoEnv.coDsp; nDsp++) {
      // Connect the DSP source pipes to the to Codec, to allow listening to the generated tones
      // from the SinewaveGenerator or the DTMF generator.
      // DSP0 => Codec0
      // DSP1 => Codec1
      // DSP2 => Codec2
      // DSP3 => Codec3
    
      result = OtxDrvCreatePipe(64, &g_hPipes.hTssCodec[nDsp]); DemoCheckRc(g_hPipes.hTssCodec[nDsp], result);
    
      memset(sqLiTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
      sqLiTimeSlotMask[g_nCodecTs] = 0xFF;
      result = OtxDrvConnectPipe(g_hPipes.hTssCodec[nDsp], OTX_PIPE_INPUT,  g_hPhysDevs.hTss, nDsp + DEMO_ASM0_HIGHWAY, sqLiTimeSlotMask, 32); DemoCheckRc(g_hPipes.hTssCodec[nDsp], result);

      nTsLi = 1+nDsp;
      memset(sqLiTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
      sqLiTimeSlotMask[nTsLi] = 0xFF;
      result = OtxDrvConnectPipe(g_hPipes.hTssCodec[nDsp], OTX_PIPE_OUTPUT, g_hPhysDevs.hTss, DEMO_CODEC_HIGHWAY, sqLiTimeSlotMask, 32); DemoCheckRc(g_hPipes.hTssCodec[nDsp], result);
    }  
  }  
  

/*
  {
    if (g_demoEnv.nBoardType != OTX_DEVICE_THOR_PCMCIA) {
      
      // Test Data Converter Device
      result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[0], OTX_LDEVICE_DATA_CONVERTER, g_hEventQueue, 0, &g_sqDataConverter[0].hHandle); DemoCheckRc(g_hPhysDevs.hDsp[0], result);
      
      result = OtxDrvCreatePipe(64, &g_sqDataConverter[0].hFromHsPipe); DemoCheckRc(g_sqDataConverter[0].hFromHsPipe, result);
      result = OtxDrvCreatePipe(64, &g_sqDataConverter[0].hToHsPipe);   DemoCheckRc(g_sqDataConverter[0].hToHsPipe, result);
      
      result = OtxDrvConnectPipe(g_sqDataConverter[0].hFromHsPipe, OTX_PIPE_INPUT,  g_hPhysDevs.hCodec[2], -1, NULL, 0); DemoCheckRc(g_sqDataConverter[0].hFromHsPipe, result);
      result = OtxDrvConnectPipe(g_sqDataConverter[0].hFromHsPipe, OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[0],   -1, NULL, 0); DemoCheckRc(g_sqDataConverter[0].hFromHsPipe, result);
      
      result = OtxDrvConnectPipe(g_sqDataConverter[0].hToHsPipe, OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[0],   -1, NULL, 0); DemoCheckRc(g_sqDataConverter[0].hToHsPipe, result);
      result = OtxDrvConnectPipe(g_sqDataConverter[0].hToHsPipe, OTX_PIPE_OUTPUT, g_hPhysDevs.hCodec[3], -1, NULL, 0); DemoCheckRc(g_sqDataConverter[0].hToHsPipe, result);
      
      result = OtxDrvConnectLogicalDevice(g_sqDataConverter[0].hFromHsPipe, -1, g_sqDataConverter[0].hHandle,      -1); DemoCheckRc(g_sqDataConverter[0].hFromHsPipe, result);
      result = OtxDrvConnectLogicalDevice(g_sqDataConverter[0].hHandle,        -1, g_sqDataConverter[0].hToHsPipe, -1); DemoCheckRc(g_sqDataConverter[0].hFromHsPipe, result);
      
      result = OtxDataConverterInit(g_sqDataConverter[0].hHandle, g_demoEnv.law, g_demoEnv.law); DemoCheckRc(g_sqDataConverter[0].hHandle, result);
      
      result = OtxDrvEnable(g_sqDataConverter[0].hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqDataConverter[0].hHandle, result);
    }
  }
*/

/*
#if 0 // One hard-coded DTMF detector connection
  nLi = 0;
  nDsp = 0;
  OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_HIGHWAY_SPEED, &nDspHighwaySpeed); 
  coTsPerDspHwy = nDspHighwaySpeed / 64;
  nTsLi = 4;
  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_EXPRESSCARD) {
    // The Li0 timeslots on this card are the TS0, TS4, TS8, ... , TS124 timeslots of the DSP highway.
    // The Li1 timeslots on this card are the TS2, TS6, TS10, ... , TS126 timeslots of the DSP highway.    
    nTsDsp = nTsLi * 4 + nLi * 2 ;  // See comment above regarding the timeslot mapping for the DSP highway on this board
  } else {
    //nTsDsp = nTsLi;
    nTsDsp = 125;
  }
  memset(sqLiTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
  sqLiTimeSlotMask[nTsLi] = 0xFF;
  memset(sqDspTimeSlotMask, 0, coTsPerDspHwy * sizeof(OTX_UINT8));
  sqDspTimeSlotMask[nTsDsp] = 0xFF;
  DemoLogf("Connecting DTMF detector device[%d]: nLi=%d, nTsLi=%.2d, nDsp=%d, nTsDsp=%.3d\n", i, nLi, nTsLi, nDsp, nTsDsp);

  OtxDrvDbgDump(g_hPhysDevs.hDsp[nDsp], 0);

  result = OtxDrvDisconnectLogicalDevice(g_hPipes.hLiDspDtmfDetector[0], -1, g_hLogicDevs.hDtmfDetector[0], -1); DemoCheckRc(g_hPipes.hLiDspDtmfDetector[0], result);
  result = OtxDrvDisconnectPipe(g_hPipes.hLiDspDtmfDetector[0], OTX_PIPE_INPUT);
  result = OtxDrvDisconnectPipe(g_hPipes.hLiDspDtmfDetector[0], OTX_PIPE_OUTPUT);

  result = OtxDrvConnectPipe(g_hPipes.hLiDspDtmfDetector[0], OTX_PIPE_INPUT, g_hPhysDevs.hLi[nLi],  -1, sqLiTimeSlotMask, 32); DemoCheckRc(g_hPipes.hLiDspDtmfDetector[0], result);
  result = OtxDrvConnectPipe(g_hPipes.hLiDspDtmfDetector[0], OTX_PIPE_OUTPUT,  g_hPhysDevs.hDsp[nDsp], -1, sqDspTimeSlotMask, coTsPerDspHwy); DemoCheckRc(g_hPipes.hLiDspDtmfDetector[0], result);
  result = OtxDrvConnectLogicalDevice(g_hPipes.hLiDspDtmfDetector[0], -1, g_hLogicDevs.hDtmfDetector[0], -1); DemoCheckRc(g_hPipes.hLiDspDtmfDetector[0], result);

  OtxDrvDbgDump(g_hPhysDevs.hDsp[nDsp], 0);
#endif
*/

}


/*******************************************************************************

  @func
    Connect Codec speaker source.

*******************************************************************************/
void DemoConnectCodecSpeakerSource(
  OTX_INT32 ts
)
{
  OTX_INT32 nDsp;
  OTX_RESULT result;
  OTX_UINT8 sqTimeSlotMask[32];

  for (nDsp = 0; nDsp < g_demoEnv.coDsp; nDsp++) {
    result = OtxDrvDisconnectPipe(g_hPipes.hTssCodec[nDsp], OTX_PIPE_INPUT); DemoCheckRc(g_hPipes.hTssCodec[nDsp], result);
    
    memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
    sqTimeSlotMask[ts] = 0xFF;
    result = OtxDrvConnectPipe(g_hPipes.hTssCodec[nDsp], OTX_PIPE_INPUT,  g_hPhysDevs.hTss, nDsp + DEMO_ASM0_HIGHWAY, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hTssCodec[nDsp], result);
    
  }        

}

/*******************************************************************************

  @func
    Initialize the logical device. This function must be called after
    the devices are created and connected.

*******************************************************************************/
void DemoInitializeLogicalDevices(
  void
)
{
  OTX_RESULT result;
  OTX_INT32  i;

  for (i = 0; i < g_demoEnv.coLi; i++) {

    // Initialize the Hdlc Sender and Receiver
    result = OtxHdlcSenderInit(
               g_hLogicDevs.hHdlcSender[i], // Handle to logical HDLC Filter Device
               OTX_HDLC_FFT_FLAGS,          // Fill character(s) to be send between HDLC frames
               2,                           // Minimum No of interframe time characters between 2 HDLC frames (1=shared flags, 2=non-shared) Min value is 1. Max is 1024.
               OTX_HDLC_CRC16,              // Cyclic Redundancy Check (CRC) mode: 16-bit, 32-bit, or none.
               OTX_FALSE,                   // If OTX_TRUE then all bits are inverted when transmitted and before received
               OTX_TRUE,                    // If OTX_TRUE the HDLC device will append the CRC to the frame to be transmitted. If OTX_FALSE, the CRC is generated with software
               OTX_HDLC_TR_NA);             // HDLC Transmission Rate for V.110 and X.30 modes
               DemoCheckRc(g_hLogicDevs.hHdlcSender[i], result);

    result = OtxHdlcReceiverInit(
               g_hLogicDevs.hHdlcReceiver[i], // Handle to logical HDLC Filter Device
               OTX_HDLC_FFT_FLAGS,          // Fill character(s) to be send between HDLC frames
               2,                           // Minimum No of interframe time characters between 2 HDLC frames (1=shared flags, 2=non-shared) Min value is 1. Max is 1024.
               OTX_HDLC_CRC16,              // Cyclic Redundancy Check (CRC) mode: 16-bit, 32-bit, or none.
               OTX_FALSE,                   // If OTX_TRUE then all bits are inverted when transmitted and before received
               OTX_TRUE,                    // If OTX_TRUE the HDLC device will remove the CRC from the received frame, if OTX_FALSE the CRC is passed on to the receiving software
               OTX_HDLC_TR_NA,              // HDLC Transmission Rate for V.110 and X.30 modes
               OTX_TRUE);                   // If OTX_TRUE the HDLC device will generate an event for every received frame, if OTX_FALSE then the HDLC device will only generate an event when one or more frames have been received
               DemoCheckRc(g_hLogicDevs.hHdlcReceiver[i], result);
  }

  for (i = 0; i < g_demoEnv.coDsp; i++) {

    if (g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM1 || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2 || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2EX || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2_INTMEM || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2EX_INTMEM || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2_EXTMEM) {
#ifdef DEMO_DUAL_TONE_DETECTOR
      OTX_UINT32    sqFreq[4];       // An array of up to 6 frequencies to be detected.
      OTX_INT32     coFreq = 4;       // Number of elements in <p sqFreq>
      
      sqFreq[0] = 941;   // in Hz
      sqFreq[1] = 1336;  // in Hz       
      sqFreq[2] = 697;   // in Hz
      sqFreq[3] = 1209;  // in Hz       
      
      
      // Initialize the Generic Dual Tone Detector. Use a couple of DTMF tone frequencies as an example
      result =  OtxToneDualDetectorInit(
        g_hLogicDevs.hDtmfDetector[i],           // Handle to the Logical device
        g_demoEnv.law,    // Data Format of the Monitored Data Stream
        sqFreq,           // An array of up to 6 frequencies to be detected.
        coFreq,           // Number of elements in <p sqFreq>
        OTX_TRUE,         // If OTX_TRUE, generate events and store the received digits
        OTX_TRUE,         // If OTX_TRUE the detector generates an event when a tone ends
        12                // Minimum duration required to validate a digit
        );
      DemoCheckRc(g_hLogicDevs.hDtmfDetector[i], result);
#else
      // Initialize the DTMF Detector
      result = OtxToneDtmfDetectorInit(g_hLogicDevs.hDtmfDetector[i], g_demoEnv.law, OTX_TRUE); DemoCheckRc(g_hLogicDevs.hDtmfDetector[i], result);
#endif





    }


    if (g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2 || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2EX || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2_INTMEM || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2EX_INTMEM || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2_EXTMEM) {
#if 0 // Include this code to get a more status messages from the Mf Detector
      if (i==0) {
        result = OtxDrvSetAttributeValueUint32(g_hLogicDevs.hSineWaveReceiver[i], OTX_ATTR_MF_DETECTOR_VERBOSE_STATUS, 1); DemoCheckRc(g_hLogicDevs.hSineWaveReceiver[i], result);
      }
#endif

      if (i==0) 
      {
        OTX_UINT32 nThreshold;

        result = OtxDrvGetAttributeValueUint32(g_hLogicDevs.hSineWaveReceiver[i], OTX_ATTR_MF_DETECTOR_THRESHOLD_TONE_ENERGY, &nThreshold); DemoCheckRc(g_hLogicDevs.hSineWaveReceiver[i], result);
        DemoLogf("MF Detector[%d]: OTX_ATTR_MF_DETECTOR_THRESHOLD_TONE_ENERGY  = %d\n", i, nThreshold);

        result = OtxDrvGetAttributeValueUint32(g_hLogicDevs.hSineWaveReceiver[i], OTX_ATTR_MF_DETECTOR_THRESHOLD_PAUSE_ENERGY, &nThreshold); DemoCheckRc(g_hLogicDevs.hSineWaveReceiver[i], result);
        DemoLogf("MF Detector[%d]: OTX_ATTR_MF_DETECTOR_THRESHOLD_PAUSE_ENERGY = %d\n", i, nThreshold);

        result = OtxDrvGetAttributeValueUint32(g_hLogicDevs.hSineWaveReceiver[i], OTX_ATTR_MF_DETECTOR_THRESHOLD_TWIST, &nThreshold); DemoCheckRc(g_hLogicDevs.hSineWaveReceiver[i], result);
        DemoLogf("MF Detector[%d]: OTX_ATTR_MF_DETECTOR_THRESHOLD_TWIST        = %d\n", i, nThreshold);

        result = OtxDrvGetAttributeValueUint32(g_hLogicDevs.hSineWaveReceiver[i], OTX_ATTR_MF_DETECTOR_THRESHOLD_RELATIVE_3RD, &nThreshold); DemoCheckRc(g_hLogicDevs.hSineWaveReceiver[i], result);
        DemoLogf("MF Detector[%d]: OTX_ATTR_MF_DETECTOR_THRESHOLD_RELATIVE_3RD = %d\n", i, nThreshold);
      }

#if 0 // To make the MF detector more stringent lower the TWIST threshold closer to 100
      {
        OTX_UINT32 nThreshold;

        nThreshold = 110;
        DemoLogf("MF Detector[%d]: Setting OTX_ATTR_MF_DETECTOR_THRESHOLD_TWIST        = %d\n", i, nThreshold);
        result = OtxDrvSetAttributeValueUint32(g_hLogicDevs.hSineWaveReceiver[i], OTX_ATTR_MF_DETECTOR_THRESHOLD_TWIST, nThreshold); DemoCheckRc(g_hLogicDevs.hSineWaveReceiver[i], result);
      }
#endif


      // Initialize the SineWave Receiver (MF Tone Detector)
      result = OtxToneMfDetectorInit(
        g_hLogicDevs.hSineWaveReceiver[i], 
        g_demoEnv.law,
        g_demoEnv.eMfToneType,  // MF Tone type (OTX_TONE_MF_R1, OTX_TONE_MF_R2_FORWARD, OTX_TONE_MF_R2_BACKWARD)
        OTX_TRUE,           // If OTX_TRUE, generate events and store the received digits in the Event QueusFifo, If OTX_FALSE Don't store. 
        OTX_FALSE,          // If OTX_TRUE the detector generates an event when a tone ends    
        25);                // Minimum duration (in milliseconds) required to validate a digit 
      DemoCheckRc(g_hLogicDevs.hSineWaveReceiver[i], result);
    } else if (g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM1) {
      // Initialize the SineWave Receiver (also a DTMF Detector)
      result = OtxToneDtmfDetectorInit(g_hLogicDevs.hSineWaveReceiver[i], g_demoEnv.law, OTX_TRUE); DemoCheckRc(g_hLogicDevs.hSineWaveReceiver[i], result);
    }

    // Initialize the DTMF Diallers
    result = OtxToneDtmfDiallerInit(
      g_hLogicDevs.hDtmfDialler[i], 
      g_demoEnv.law, 
      8, 
      8, 
      OTX_TONE_GAIN_0DB); 
    DemoCheckRc(g_hLogicDevs.hDtmfDialler[i], result);

    if (g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2 || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2EX || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2_INTMEM || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2EX_INTMEM || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2_EXTMEM) {
      // Initialize the MF Diallers
      result = OtxToneMfDiallerInit(
        g_hLogicDevs.hMfDialler[i], 
        g_demoEnv.law, 
        g_demoEnv.eMfToneType,
        50, 
        25, 
        OTX_TONE_GAIN_0DB); 
      DemoCheckRc(g_hLogicDevs.hMfDialler[i], result);
    }


  }

}


/*******************************************************************************

  @func
    Enable Configured Logical Devices

*******************************************************************************/
void DemoEnableLogicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  for (i = 0; i < g_demoEnv.coLi; i++) {
    result = OtxDrvEnable(g_hLogicDevs.hHdlcSender[i],   OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hHdlcSender[i], result);
    result = OtxDrvEnable(g_hLogicDevs.hHdlcReceiver[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hHdlcReceiver[i], result);
  }

  for (i = 0; i < g_demoEnv.coDsp; i++) {
    if (g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM1 || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2 || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2EX || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2_INTMEM || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2EX_INTMEM || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2_EXTMEM) {
      result = OtxDrvEnable(g_hLogicDevs.hDtmfDetector[i],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hDtmfDetector[i], result);
      result = OtxDrvEnable(g_hLogicDevs.hSineWaveReceiver[i],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hSineWaveReceiver[i], result);
    }

    if (g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2 || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2EX || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2_INTMEM || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2EX_INTMEM || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2_EXTMEM) {
      result = OtxDrvEnable(g_hLogicDevs.hMfDialler[i],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hMfDialler[i], result);
    }
    result = OtxDrvEnable(g_hLogicDevs.hDtmfDialler[i],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hDtmfDialler[i], result);
  }

}


/*******************************************************************************

  @func
    Initialize the T1/E1 Configuration Data Structures passed to the T1/E1
    Line Interface physical devices

*******************************************************************************/
void DemoInitT1E1ConfigData(
  OUT OtxE1ConfigOptionsS *pE1CfgData,
  OUT OtxT1ConfigOptionsS *pT1CfgData
)
{
  // Initialize E1 Configuration Data
  pE1CfgData->eTxLineCode                = OTX_E1_LC_HDB3;
  pE1CfgData->eRxLineCode                = OTX_E1_LC_HDB3;
  pE1CfgData->eTxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eRxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eAisDetection              = OTX_E1_AIS_ETS300233;
  pE1CfgData->bfSiBits                   = 0x3;
  pE1CfgData->bfSaBits                   = 0x1F;
  pE1CfgData->bExtendedHDB3errorDetect   = OTX_FALSE;
  pE1CfgData->bAutoRegainMultiframe      = OTX_FALSE;
  pE1CfgData->bAutoResync                = OTX_TRUE;
  pE1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pE1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pE1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775
  pE1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775
  pE1CfgData->nLineLength                = 10;
  pE1CfgData->bHighTxPower               = OTX_FALSE;
  pE1CfgData->bReceiveEqualizer          = g_demoEnv.bReceiveEqualizer;;

  // Initialize T1 Configuration Data
  pT1CfgData->eSignalingMode             = OTX_T1_SM_CCS;
  pT1CfgData->eTxLineCode                = OTX_T1_LC_B8ZS;
  pT1CfgData->eRxLineCode                = OTX_T1_LC_B8ZS;
  pT1CfgData->eFrameFormat               = OTX_T1_FF_ESF;
  pT1CfgData->bCRC6Enable                = OTX_TRUE;
  pT1CfgData->eTxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->eRxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->bAutoResync                = OTX_TRUE;
  pT1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pT1CfgData->eLfaSensitivity            = OTX_T1_LFA_2_OUT_OF_4;
  pT1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pT1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLineLength                = 10;
  pT1CfgData->bHighTxPower               = OTX_FALSE;
  pT1CfgData->bReceiveEqualizer          = g_demoEnv.bReceiveEqualizer;;
  pT1CfgData->bfClearChannels            = 0x000000; // 0x000000 = No channels are marked as clear channels
}




void *DemoSendHdlcContinously(
  void *lpParameter   // thread data
)
{
  OTX_RESULT result;
  OTX_UINT8 arMsg[64];
  OTX_INT32 nMsgLen = 3;
  OTX_INT32 i;
  OTX_INT32 nLi = 0;
  OTX_INT32 busyCount = 0;
    
  DemoLogf("Sending random HDLC frames continously from Li #%d (until 'a' is pressed again) ...", nLi);
  
  do {  // Send HDLC Frames until user presses 'f' again

    nMsgLen = (rand() % 20) + 3;

    for (i = 0; i < nMsgLen; i++) {
      arMsg[i] = (OTX_UINT8) rand(); 
    }
    
    result = OtxHdlcSendData(g_hLogicDevs.hHdlcSender[nLi], arMsg, nMsgLen, DEMO_TASK_REF, OTX_TIME_INFINITY);
    if (OTX_RESULT_CODE(result) == OTX_E_BUSY) {
      // Driver IO Request buffer is full. Wait a while before trying to send more frames
      busyCount++;
      OsaThreadSleep(100);
    } else {
      DemoCheckRc(g_hLogicDevs.hHdlcSender[nLi], result);
    }
    
  } while (g_bContinousHdlcSendingActive);

  DemoLogf("DemoSendHdlcContinously() thread terminated (busyCount=%d)\n", busyCount);

  return 0;
}



/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*********************************************************************************/
void *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OtxEventDataS eventData;
  OTX_RESULT result;


  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);
    //DemoLogf("Event signalled\n");
    do {
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(g_hEventQueue, result);

      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event
        // Printf the event data.
        // In a real application branch based on the data
        DemoPrintEvent(&eventData);

      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);
  }

  return 0;


}



/*******************************************************************************
  @func
    Prints a Driver Event. If HDLC Data, reads the data.
*******************************************************************************/
void DemoPrintEvent(
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  OTX_UINT32 i;
  OTX_UINT8  buf[1024];
  OTX_UINT32 nFrameLen;
  OTX_UINT32 nFrameSeq;
  OtxHdlcFrameEndCodeE nFrameEndCode;
  OTX_DATETIME nTimeStamp;
  char msgStr[256];

  DemoLogf("%.8X%.8X ", pEventData->m_nTimeStamp.sUint32.nHigh, pEventData->m_nTimeStamp.sUint32.nLow);

  switch (pEventData->m_eDeviceType) {
    case OTX_DEVICE_LI_T1E1:
    // Event from Li0 or Li1
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      if (pEventData->m_nParam) {
        DemoLogf("E1/T1[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
      } else {
        DemoLogf("E1/T1[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("E1/T1[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_DEVICE_DSP:
    // Non-logical device related event from the DSPs
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("DSP[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
    } else {
      DemoLogf("DSP[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_LDEVICE_TONE_DTMF_DIALER:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("DtmfDialler[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
    } else {
      DemoLogf("DtmfDialler[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_LDEVICE_TONE_MF_DIALER:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("MfDialler[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
    } else {
      DemoLogf("MfDialler[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_LDEVICE_TONE_DTMF_DETECTOR:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      // Event from Dtmf Detector
      if (pEventData->m_nCode == OTX_DTMF_DETECTOR_EC_DIGIT) {
        switch (pEventData->m_nParam) {
        case 0:  // 0  (941Hz/1336Hz)
        case 1:  // 1  (697Hz/1209Hz)
        case 2:  // 2  (697Hz/1336Hz)
        case 3:  // 3  (697Hz/1477Hz)
        case 4:  // 4  (770Hz/1209Hz)
        case 5:  // 5  (770Hz/1336Hz)
        case 6:  // 6  (770Hz/1477Hz)
        case 7:  // 7  (852Hz/1209Hz)
        case 8:  // 8  (852Hz/1336Hz)
        case 9:  // 9  (852Hz/1477Hz)
          DemoLogf("DtmfDetector[%d] detected DTMF digit %d \n", pEventData->m_nSourceId, pEventData->m_nParam);
          break;
        case 10: // A  (697Hz/1633Hz)
        case 11: // B  (770Hz/1633Hz)
        case 12: // C  (852Hz/1633Hz)
        case 13: // D  (941Hz/1633Hz)
          DemoLogf("DtmfDetector[%d] detected DTMF key %C \n", pEventData->m_nSourceId, pEventData->m_nParam - 10 + 'A');
          break;
        case 14: // *  (941Hz/1209Hz)
          DemoLogf("DtmfDetector[%d] detected DTMF key '*'\n", pEventData->m_nSourceId);
          break;
        case 15: // #  (941Hz/1477Hz)
          DemoLogf("DtmfDetector[%d] detected DTMF key '#'\n", pEventData->m_nSourceId);
          break;
        default: 
          DemoLogf("DtmfDetector[%d] Unrecognized DTMF tone (%d)\n", pEventData->m_nSourceId, pEventData->m_nParam);
          break;
        }

      } else {
        // Unexpected event from this device
        DemoLogf("DtmfDetector[%d] Unexpected Event: %s \n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("DtmfDetector[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_LDEVICE_TONE_DUAL_DETECTOR:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      // Event from Dtmf Detector
      if (pEventData->m_nCode == OTX_DUAL_TONE_DETECTOR_EC_TONE) {
        DemoLogf("DualDetector[%d] detected Dual Tone %d \n", pEventData->m_nSourceId, pEventData->m_nParam);
      } else if (pEventData->m_nCode == OTX_DUAL_TONE_DETECTOR_EC_PAUSE) {
        DemoLogf("DualDetector[%d] detected Pause %d \n", pEventData->m_nSourceId, pEventData->m_nParam);
      } else if (pEventData->m_nCode == OTX_DUAL_TONE_DETECTOR_EC_VERBOSE_STATUS) {
        DemoLogf("DualDetector[%d] Verbose Status 0x%.8X (Code: %d, Value: %d)\n", pEventData->m_nSourceId, pEventData->m_nParam, pEventData->m_nParam & 0x0000FFFF, pEventData->m_nParam >> 16);
      } else {
        // Unexpected event from this device
        DemoLogf("DualDetector[%d] Unexpected Event: %s \n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("DualDetector[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_LDEVICE_TONE_MF_DETECTOR:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      // Event from Dtmf Detector
      if (pEventData->m_nCode == OTX_MF_DETECTOR_EC_DIGIT) {
        DemoLogf("MfDetector[%d] detected MF digit %.2d \n", pEventData->m_nSourceId, pEventData->m_nParam);
      } else if (pEventData->m_nCode == OTX_MF_DETECTOR_EC_VERBOSE_STATUS){
        DemoLogf("MfDetector[%d] Verbose Status 0x%.8X (Code: %d, Value: %d)\n", pEventData->m_nSourceId, pEventData->m_nParam, pEventData->m_nParam & 0x0000FFFF, pEventData->m_nParam >> 16);
      } else {
        if (pEventData->m_nParam) {
          DemoLogf("MfDetector[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
        } else {
          DemoLogf("MfDetector[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
        }
      }
    } else {
      DemoLogf("MfDetector[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_LDEVICE_HDLC_RECEIVER:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      // Event from HLDC Receiver
      // @todo DSP and LI HdlcReceivers should return the same event code since they are of the same logical device type
      if (pEventData->m_nCode == OTX_HDLC_RECEIVER_EC_COMPLETE_FRAME) {

        DemoLogf("HDLC Frame Received, Sequence Number = %d\n", pEventData->m_nParam);

        otxRc = OtxHdlcReadData(pEventData->m_hDevice, buf, 1024, &nFrameLen, &nFrameSeq, &nFrameEndCode, &nTimeStamp);
        if (OTX_RESULT_CODE(otxRc) == OTX_S_OK) {
          DemoLogf("HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, nFrameLen, nFrameEndCode, nFrameSeq);
          for (i = 0; i < nFrameLen && i < 1024; i++) {
            DemoLogf("%.2X ", buf[i]);
          }
          DemoLogf("\n");
        } else {
          DemoLogf("HdlcReceiver[%d] OtxHdlcReadData() was unsuccessful.\n", pEventData->m_nSourceId);
          DemoCheckRc(pEventData->m_hDevice, otxRc);
        }
      } else {
        // Unexpected event from this device
        DemoLogf("HdlcReceiver[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("HdlcReceiver[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_LDEVICE_HDLC_SENDER:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    DemoLogf("HdlcSender[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
    break;

  default:
    // Unhandled event
    DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }


}

/*******************************************************************************

  @func

  Uses the SineWaveGenerator device to send a tone (a DTMF tone).

*******************************************************************************/
void DemoSendSineWaveDtmfTone(OTX_INT32 nDialler)
{
  OTX_RESULT result;
  static OTX_INT32 nDtmfDigit = 0;

  // Initialize the SineWave Sender 
  if (g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM1 || 
      g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2 || 
      g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2EX || 
      g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2_INTMEM || 
      g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2EX_INTMEM || 
      g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2_EXTMEM) {
    OtxToneSineWaveDataS sqSineWaveData[5];

    sqSineWaveData[0].nAttenuation = -9; // -9dB is about 25% of maximum amplitude
    sqSineWaveData[0].nPhase       = 0;  

    sqSineWaveData[1].nAttenuation = -9; // -9dB is about 25% of maximum amplitude
    sqSineWaveData[1].nPhase       = 0;  

    switch (nDtmfDigit) {
    case 0:
      // DTMF digit 0 
      sqSineWaveData[0].nFrequency   = 941; // in Hz
      sqSineWaveData[1].nFrequency   = 1336; // in Hz       
      break;
    case 1:
      // DTMF digit 1 
      sqSineWaveData[0].nFrequency   = 697; // in Hz
      sqSineWaveData[1].nFrequency   = 1209; // in Hz       
      break;
    case 2:
      // DTMF digit 2 
      sqSineWaveData[0].nFrequency   = 697; // in Hz
      sqSineWaveData[1].nFrequency   = 1336; // in Hz       
      break;
    case 3:
      // DTMF digit 3 
      sqSineWaveData[0].nFrequency   = 697; // in Hz
      sqSineWaveData[1].nFrequency   = 1477; // in Hz       
      break;
    case 4:
      // DTMF digit 4 
      sqSineWaveData[0].nFrequency   = 770; // in Hz
      sqSineWaveData[1].nFrequency   = 1209; // in Hz       
      break;
    case 5:
      // DTMF digit 5 
      sqSineWaveData[0].nFrequency   = 770; // in Hz
      sqSineWaveData[1].nFrequency   = 1336; // in Hz       
      break;
    case 6:
      // DTMF digit 6
      sqSineWaveData[0].nFrequency   = 770; // in Hz
      sqSineWaveData[1].nFrequency   = 1477; // in Hz       
      break;
    case 7:
      // DTMF digit 7 
      sqSineWaveData[0].nFrequency   = 852; // in Hz
      sqSineWaveData[1].nFrequency   = 1209; // in Hz       
      break;
    case 8:
      // DTMF digit 8 
      sqSineWaveData[0].nFrequency   = 852; // in Hz
      sqSineWaveData[1].nFrequency   = 1336; // in Hz       
      break;
    case 9:
      // DTMF digit 9
      sqSineWaveData[0].nFrequency   = 852; // in Hz
      sqSineWaveData[1].nFrequency   = 1477; // in Hz       
      break;
    }
    
    DemoLogf("Initalizing SineWave Generator[%d] for DTMF tone %d (%d Hz / %d Hz)\n", nDialler, nDtmfDigit, (OTX_INT32)sqSineWaveData[0].nFrequency, (OTX_INT32)sqSineWaveData[1].nFrequency);
    result = OtxToneSineWaveGeneratorInit(g_hLogicDevs.hSineWaveSender[nDialler], 
      g_demoEnv.law, 
      sqSineWaveData,
      2, // coSineWaveData
      OTX_TONE_GAIN_0DB); DemoCheckRc(g_hLogicDevs.hSineWaveSender[nDialler], result);
    
    
    DemoLogf("Enabling SineWave Generator[%d]... \n", nDialler);
    result = OtxDrvEnable(g_hLogicDevs.hSineWaveSender[nDialler],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hSineWaveSender[nDialler], result);
    
    OsaThreadSleep(500);
    
    DemoLogf("Disabling SineWave Generator[%d]... \n", nDialler);
    result = OtxDrvDisable(g_hLogicDevs.hSineWaveSender[nDialler],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hSineWaveSender[nDialler], result);

    if (++nDtmfDigit > 9) {
      nDtmfDigit = 0;
    }

    
  } else {
    DemoLogf("Function not supported for this SPM.\n");
  }
  

}


/*******************************************************************************

  @func

  Uses the SineWaveGenerator device to send a tone (a MF T1 R1).

*******************************************************************************/
void DemoSendSineWaveMfR1T1Tone(OTX_INT32 nDialler)
{
  OTX_RESULT result;
  static OTX_INT32 nMfDigit = 0;
  char *szMfDigitStr = "?";

  // Initialize the SineWave Sender 
  if (g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM1 || 
      g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2 || 
      g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2EX || 
      g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2_INTMEM || 
      g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2EX_INTMEM || 
      g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2_EXTMEM) {
    OtxToneSineWaveDataS sqSineWaveData[5];

    sqSineWaveData[0].nAttenuation = -9; // -9dB is about 25% of maximum amplitude
    sqSineWaveData[0].nPhase       = 0;  

    sqSineWaveData[1].nAttenuation = -9; // -9dB is about 25% of maximum amplitude
    sqSineWaveData[1].nPhase       = 0;  

    switch (nMfDigit) {
    case 0:
      // MF digit 0 
      szMfDigitStr = "0";
      sqSineWaveData[0].nFrequency   = 1300; // in Hz
      sqSineWaveData[1].nFrequency   = 1500; // in Hz       
      break;
    case 1:
      // MF digit 1 
      szMfDigitStr = "1";
      sqSineWaveData[0].nFrequency   = 700; // in Hz
      sqSineWaveData[1].nFrequency   = 900; // in Hz       
      break;
    case 2:
      // MF digit 2 
      szMfDigitStr = "2";
      sqSineWaveData[0].nFrequency   = 700; // in Hz
      sqSineWaveData[1].nFrequency   = 1100; // in Hz       
      break;
    case 3:
      // MF digit 3 
      szMfDigitStr = "3";
      sqSineWaveData[0].nFrequency   = 900; // in Hz
      sqSineWaveData[1].nFrequency   = 1100; // in Hz       
      break;
    case 4:
      // MF digit 4 
      szMfDigitStr = "4";
      sqSineWaveData[0].nFrequency   = 700; // in Hz
      sqSineWaveData[1].nFrequency   = 1300; // in Hz       
      break;
    case 5:
      // MF digit 5 
      szMfDigitStr = "5";
      sqSineWaveData[0].nFrequency   = 900; // in Hz
      sqSineWaveData[1].nFrequency   = 1300; // in Hz       
      break;
    case 6:
      // MF digit 6
      szMfDigitStr = "6";
      sqSineWaveData[0].nFrequency   = 1100; // in Hz
      sqSineWaveData[1].nFrequency   = 1300; // in Hz       
      break;
    case 7:
      // MF digit 7 
      szMfDigitStr = "7";
      sqSineWaveData[0].nFrequency   = 700; // in Hz
      sqSineWaveData[1].nFrequency   = 1500; // in Hz       
      break;
    case 8:
      // MF digit 8 
      szMfDigitStr = "8";
      sqSineWaveData[0].nFrequency   = 900; // in Hz
      sqSineWaveData[1].nFrequency   = 1500; // in Hz       
      break;
    case 9:
      // MF digit 9
      szMfDigitStr = "9";
      sqSineWaveData[0].nFrequency   = 1100; // in Hz
      sqSineWaveData[1].nFrequency   = 1500; // in Hz       
      break;
    case 10:
      // MF KP
      szMfDigitStr = "KP(10)";
      sqSineWaveData[0].nFrequency   = 1100; // in Hz
      sqSineWaveData[1].nFrequency   = 1700; // in Hz       
      break;
    case 11:
      // MF ST
      szMfDigitStr = "ST(11)";
      sqSineWaveData[0].nFrequency   = 1500; // in Hz
      sqSineWaveData[1].nFrequency   = 1700; // in Hz       
      break;
    case 12:
      // MF ST1(x)
      szMfDigitStr = "ST1(12)";
      sqSineWaveData[0].nFrequency   = 900; // in Hz
      sqSineWaveData[1].nFrequency   = 1700; // in Hz       
      break;
    case 13:
      // MF ST2(y)
      szMfDigitStr = "ST2(13)";
      sqSineWaveData[0].nFrequency   = 1300; // in Hz
      sqSineWaveData[1].nFrequency   = 1700; // in Hz       
      break;
    case 14:
      // MF ST3(z)
      szMfDigitStr = "ST3(14)";
      sqSineWaveData[0].nFrequency   = 700; // in Hz
      sqSineWaveData[1].nFrequency   = 1700; // in Hz       
      break;
    }
    
    DemoLogf("Initalizing SineWave Generator[%d] for R1 MF tone %s (%d Hz / %d Hz)\n", nDialler, szMfDigitStr, (OTX_INT32)sqSineWaveData[0].nFrequency, (OTX_INT32)sqSineWaveData[1].nFrequency);
    result = OtxToneSineWaveGeneratorInit(g_hLogicDevs.hSineWaveSender[nDialler], 
      g_demoEnv.law, 
      sqSineWaveData,
      2, // coSineWaveData
      OTX_TONE_GAIN_0DB); DemoCheckRc(g_hLogicDevs.hSineWaveSender[nDialler], result);
    
    
    DemoLogf("Enabling SineWave Generator[%d]... \n", nDialler);
    result = OtxDrvEnable(g_hLogicDevs.hSineWaveSender[nDialler],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hSineWaveSender[nDialler], result);
    
    OsaThreadSleep(500);
    
    DemoLogf("Disabling SineWave Generator[%d]... \n", nDialler);
    result = OtxDrvDisable(g_hLogicDevs.hSineWaveSender[nDialler],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hSineWaveSender[nDialler], result);

    if (++nMfDigit > 14) {
      nMfDigit = 0;
    }

    
  } else {
    DemoLogf("Function not supported for this SPM.\n");
  }
  

}


/*******************************************************************************

  @func

  Uses the SineWaveGenerator device to send a tone (a MF E1 R2 Forward).

*******************************************************************************/
void DemoSendSineWaveMfR2E1ForwardTone(OTX_INT32 nDialler)
{
  OTX_RESULT result;
  static OTX_INT32 nMfDigit = 1;
  char *szMfDigitStr = "?";

  // Initialize the SineWave Sender 
  if (g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM1 || 
      g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2 || 
      g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2EX || 
      g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2_INTMEM || 
      g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2EX_INTMEM || 
      g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2_EXTMEM) {
    OtxToneSineWaveDataS sqSineWaveData[5];

    sqSineWaveData[0].nAttenuation = -9; // -9dB is about 25% of maximum amplitude
    sqSineWaveData[0].nPhase       = 0;  

    sqSineWaveData[1].nAttenuation = -9; // -9dB is about 25% of maximum amplitude
    sqSineWaveData[1].nPhase       = 0;  

    switch (nMfDigit) {
    case 1:
      // MF digit 1 
      szMfDigitStr = "1";
      sqSineWaveData[0].nFrequency   = 1380; // in Hz
      sqSineWaveData[1].nFrequency   = 1500; // in Hz       
      break;
    case 2:
      // MF digit 2 
      szMfDigitStr = "2";
      sqSineWaveData[0].nFrequency   = 1380; // in Hz
      sqSineWaveData[1].nFrequency   = 1620; // in Hz       
      break;
    case 3:
      // MF digit 3 
      szMfDigitStr = "3";
      sqSineWaveData[0].nFrequency   = 1500; // in Hz
      sqSineWaveData[1].nFrequency   = 1620; // in Hz       
      break;
    case 4:
      // MF digit 4 
      szMfDigitStr = "4";
      sqSineWaveData[0].nFrequency   = 1380; // in Hz
      sqSineWaveData[1].nFrequency   = 1740; // in Hz       
      break;
    case 5:
      // MF digit 5 
      szMfDigitStr = "5";
      sqSineWaveData[0].nFrequency   = 1500; // in Hz
      sqSineWaveData[1].nFrequency   = 1740; // in Hz       
      break;
    case 6:
      // MF digit 6
      szMfDigitStr = "6";
      sqSineWaveData[0].nFrequency   = 1620; // in Hz
      sqSineWaveData[1].nFrequency   = 1740; // in Hz       
      break;
    case 7:
      // MF digit 7 
      szMfDigitStr = "7";
      sqSineWaveData[0].nFrequency   = 1380; // in Hz
      sqSineWaveData[1].nFrequency   = 1860; // in Hz       
      break;
    case 8:
      // MF digit 8 
      szMfDigitStr = "8";
      sqSineWaveData[0].nFrequency   = 1500; // in Hz
      sqSineWaveData[1].nFrequency   = 1860; // in Hz       
      break;
    case 9:
      // MF digit 9
      szMfDigitStr = "9";
      sqSineWaveData[0].nFrequency   = 1620; // in Hz
      sqSineWaveData[1].nFrequency   = 1860; // in Hz       
      break;
    case 10:
      // MF digit a
      szMfDigitStr = "10";
      sqSineWaveData[0].nFrequency   = 1740; // in Hz
      sqSineWaveData[1].nFrequency   = 1860; // in Hz       
      break;
    case 11:
      // MF digit b
      szMfDigitStr = "11";
      sqSineWaveData[0].nFrequency   = 1380; // in Hz
      sqSineWaveData[1].nFrequency   = 1980; // in Hz       
      break;
    case 12:
      // MF digit c
      szMfDigitStr = "12";
      sqSineWaveData[0].nFrequency   = 1500; // in Hz
      sqSineWaveData[1].nFrequency   = 1980; // in Hz       
      break;
    case 13:
      // MF digit d
      szMfDigitStr = "13";
      sqSineWaveData[0].nFrequency   = 1620; // in Hz
      sqSineWaveData[1].nFrequency   = 1980; // in Hz       
      break;
    case 14:
      // MF digit e
      szMfDigitStr = "14";
      sqSineWaveData[0].nFrequency   = 1740; // in Hz
      sqSineWaveData[1].nFrequency   = 1980; // in Hz       
      break;
    case 15:
      // MF digit f
      szMfDigitStr = "15";
      sqSineWaveData[0].nFrequency   = 1860; // in Hz
      sqSineWaveData[1].nFrequency   = 1980; // in Hz       
      break;
    }
    
    DemoLogf("Initalizing SineWave Generator[%d] for R2 MF Forward tone %s (%d Hz / %d Hz)\n", nDialler, szMfDigitStr, (OTX_INT32)sqSineWaveData[0].nFrequency, (OTX_INT32)sqSineWaveData[1].nFrequency);
    result = OtxToneSineWaveGeneratorInit(g_hLogicDevs.hSineWaveSender[nDialler], 
      g_demoEnv.law, 
      sqSineWaveData,
      2, // coSineWaveData
      OTX_TONE_GAIN_0DB); DemoCheckRc(g_hLogicDevs.hSineWaveSender[nDialler], result);
    
    
    DemoLogf("Enabling SineWave Generator[%d]... \n", nDialler);
    result = OtxDrvEnable(g_hLogicDevs.hSineWaveSender[nDialler],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hSineWaveSender[nDialler], result);
    
    OsaThreadSleep(250);
    //OtxDrvDbgDump(g_hPhysDevs.hDsp[1], 0);
    OsaThreadSleep(250);
    
    DemoLogf("Disabling SineWave Generator[%d]... \n", nDialler);
    result = OtxDrvDisable(g_hLogicDevs.hSineWaveSender[nDialler],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hSineWaveSender[nDialler], result);

    if (++nMfDigit > 15) {
      nMfDigit = 1;
    }

    OsaThreadSleep(100);
    //OtxDrvDbgDump(g_hPhysDevs.hDsp[1], 0);
    
  } else {
    DemoLogf("Function not supported for this SPM.\n");
  }
  

}


/*******************************************************************************

  @func

  Uses the SineWaveGenerator device to send a tone (a MF E1 R2 Backward).

*******************************************************************************/
void DemoSendSineWaveMfR2E1BackwardTone(OTX_INT32 nDialler)
{
  OTX_RESULT result;
  static OTX_INT32 nMfDigit = 1;
  char *szMfDigitStr = "?";

  // Initialize the SineWave Sender 
  if (g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM1 || 
      g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2 || 
      g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2EX || 
      g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2_INTMEM || 
      g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2EX_INTMEM || 
      g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2_EXTMEM) {
    OtxToneSineWaveDataS sqSineWaveData[5];

    sqSineWaveData[0].nAttenuation = -9; // -9dB is about 25% of maximum amplitude
    sqSineWaveData[0].nPhase       = 0;  

    sqSineWaveData[1].nAttenuation = -9; // -9dB is about 25% of maximum amplitude
    sqSineWaveData[1].nPhase       = 0;  

    switch (nMfDigit) {
    case 1:
      // MF digit 1 
      szMfDigitStr = "1";
      sqSineWaveData[0].nFrequency   = 1140; // in Hz
      sqSineWaveData[1].nFrequency   = 1020; // in Hz       
      break;
    case 2:
      // MF digit 2 
      szMfDigitStr = "2";
      sqSineWaveData[0].nFrequency   = 1140; // in Hz
      sqSineWaveData[1].nFrequency   = 900; // in Hz       
      break;
    case 3:
      // MF digit 3 
      szMfDigitStr = "3";
      sqSineWaveData[0].nFrequency   = 1020; // in Hz
      sqSineWaveData[1].nFrequency   = 900; // in Hz       
      break;
    case 4:
      // MF digit 4 
      szMfDigitStr = "4";
      sqSineWaveData[0].nFrequency   = 1140; // in Hz
      sqSineWaveData[1].nFrequency   = 780; // in Hz       
      break;
    case 5:
      // MF digit 5 
      szMfDigitStr = "5";
      sqSineWaveData[0].nFrequency   = 1020; // in Hz
      sqSineWaveData[1].nFrequency   = 780; // in Hz       
      break;
    case 6:
      // MF digit 6
      szMfDigitStr = "6";
      sqSineWaveData[0].nFrequency   = 900; // in Hz
      sqSineWaveData[1].nFrequency   = 780; // in Hz       
      break;
    case 7:
      // MF digit 7 
      szMfDigitStr = "7";
      sqSineWaveData[0].nFrequency   = 1140; // in Hz
      sqSineWaveData[1].nFrequency   = 660; // in Hz       
      break;
    case 8:
      // MF digit 8 
      szMfDigitStr = "8";
      sqSineWaveData[0].nFrequency   = 1020; // in Hz
      sqSineWaveData[1].nFrequency   = 660; // in Hz       
      break;
    case 9:
      // MF digit 9
      szMfDigitStr = "9";
      sqSineWaveData[0].nFrequency   = 900; // in Hz
      sqSineWaveData[1].nFrequency   = 660; // in Hz       
      break;
    case 10:
      // MF digit a
      szMfDigitStr = "10";
      sqSineWaveData[0].nFrequency   = 780; // in Hz
      sqSineWaveData[1].nFrequency   = 660; // in Hz       
      break;
    case 11:
      // MF digit b
      szMfDigitStr = "11";
      sqSineWaveData[0].nFrequency   = 1140; // in Hz
      sqSineWaveData[1].nFrequency   = 540; // in Hz       
      break;
    case 12:
      // MF digit c
      szMfDigitStr = "12";
      sqSineWaveData[0].nFrequency   = 1020; // in Hz
      sqSineWaveData[1].nFrequency   = 540; // in Hz       
      break;
    case 13:
      // MF digit d
      szMfDigitStr = "13";
      sqSineWaveData[0].nFrequency   = 900; // in Hz
      sqSineWaveData[1].nFrequency   = 540; // in Hz       
      break;
    case 14:
      // MF digit e
      szMfDigitStr = "14";
      sqSineWaveData[0].nFrequency   = 780; // in Hz
      sqSineWaveData[1].nFrequency   = 540; // in Hz       
      break;
    case 15:
      // MF digit f
      szMfDigitStr = "15";
      sqSineWaveData[0].nFrequency   = 660; // in Hz
      sqSineWaveData[1].nFrequency   = 540; // in Hz       
      break;
    }
    
    DemoLogf("Initalizing SineWave Generator[%d] for R2 MF Backward tone %s (%d Hz / %d Hz)\n", nDialler, szMfDigitStr, (OTX_INT32)sqSineWaveData[0].nFrequency, (OTX_INT32)sqSineWaveData[1].nFrequency);
    result = OtxToneSineWaveGeneratorInit(g_hLogicDevs.hSineWaveSender[nDialler], 
      g_demoEnv.law, 
      sqSineWaveData,
      2, // coSineWaveData
      OTX_TONE_GAIN_0DB); DemoCheckRc(g_hLogicDevs.hSineWaveSender[nDialler], result);
    
    
    DemoLogf("Enabling SineWave Generator[%d]... \n", nDialler);
    result = OtxDrvEnable(g_hLogicDevs.hSineWaveSender[nDialler],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hSineWaveSender[nDialler], result);
    
    OsaThreadSleep(500);
    
    DemoLogf("Disabling SineWave Generator[%d]... \n", nDialler);
    result = OtxDrvDisable(g_hLogicDevs.hSineWaveSender[nDialler],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hSineWaveSender[nDialler], result);

    if (++nMfDigit > 15) {
      nMfDigit = 1;
    }

    
  } else {
    DemoLogf("Function not supported for this SPM.\n");
  }
  

}


/*******************************************************************************

  @func

  Uses the SineWaveGenerator device to send a generic tone. Prompts for 
  frequencies, attenuation and duration.

*******************************************************************************/
void DemoSendGenericSineWave(OTX_INT32 nDialler)
{
  OTX_RESULT result;
  OTX_INT32 i;
  static OTX_INT32 nSineWaveComponents = 1;
  static OTX_INT32 nFrequency = 1000;
  static OTX_INT32 nAttenuation = 0;  // <=0
  static OTX_INT32 nDuration = 1000;  // in milliseconds
  char strBuf[64];

  // Initialize the SineWave Sender 
  if (g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM1 || 
      g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2 || 
      g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2EX || 
      g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2_INTMEM || 
      g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2EX_INTMEM || 
      g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2_EXTMEM) {
    OtxToneSineWaveDataS sqSineWaveData[5];

    DemoLogf("Please enter duration generated tone (in milliseconds) [%d]", nDuration);
    OsaConsoleGetStr(strBuf);
    if (*strBuf) {
      nDuration = strtol(strBuf, NULL, 0);
    }

    if (nDuration < 1) {
      DemoLogf("Invalid range. Operation aborted.\n");
      return;
    }
      
    DemoLogf("Please enter number of frequencies in the generated tone (1-5) [%d]", nSineWaveComponents);
    OsaConsoleGetStr(strBuf);
    if (*strBuf) {
      nSineWaveComponents = strtol(strBuf, NULL, 0);
    }

    if ((nSineWaveComponents < 1) || (nSineWaveComponents > 5)) {
      DemoLogf("Invalid range. Operation aborted.\n");
      return;
    }


    for (i = 0; i < nSineWaveComponents; i++) {

      DemoLogf("[%d]: Frequency? (1-8000) [%d]", i, nFrequency);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        nFrequency = strtol(strBuf, NULL, 0);
      }
      if ((nFrequency < 1) || (nFrequency > 8000)) {
        DemoLogf("Invalid range. Operation aborted.\n");
        return;
      }
      
      DemoLogf("[%d]: Attenuation? (<=0) [%+d]", i, nAttenuation);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        nAttenuation = strtol(strBuf, NULL, 0);
      }
      if (nAttenuation > 0) {
        DemoLogf("Invalid range. Operation aborted.\n");
        return;
      }
      
      sqSineWaveData[i].nFrequency   = nFrequency; // in Hz
      sqSineWaveData[i].nAttenuation = nAttenuation; 
      sqSineWaveData[i].nPhase       = 0;  
    }
    
    DemoLogf("Initalizing SineWave Generator[%d] for tone %d", nDialler, (OTX_INT32)sqSineWaveData[0].nFrequency);
    for (i = 0; i < nSineWaveComponents; i++) {
      DemoLogf("/ %d ", (OTX_INT32)sqSineWaveData[i].nFrequency);
    }
    DemoLogf(" Hz. Duration = %d ms\n", nDuration);

    DemoLogf("Press <Enter> to start tone generation...\n");
    OsaConsoleGetStr(strBuf);

    result = OtxToneSineWaveGeneratorInit(g_hLogicDevs.hSineWaveSender[nDialler], 
      g_demoEnv.law, 
      sqSineWaveData,
      nSineWaveComponents, // coSineWaveData
      OTX_TONE_GAIN_0DB); DemoCheckRc(g_hLogicDevs.hSineWaveSender[nDialler], result);
    
    
    DemoLogf("Enabling SineWave Generator[%d]... \n", nDialler);
    result = OtxDrvEnable(g_hLogicDevs.hSineWaveSender[nDialler],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hSineWaveSender[nDialler], result);
    
    OsaThreadSleep(nDuration);
    
    DemoLogf("Disabling SineWave Generator[%d]... \n", nDialler);
    result = OtxDrvDisable(g_hLogicDevs.hSineWaveSender[nDialler],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hSineWaveSender[nDialler], result);

    
  } else {
    DemoLogf("Function not supported for this SPM.\n");
  }
  

}




/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) {
  short  i;

  // Set reasonable defaults before reading from command line
  envP->pszInputStr      = NULL;
  envP->bExitOnError     = OTX_FALSE;
  envP->law              = OTX_DATA_FORMAT_COMPAND8_A_LAW;
  envP->eT1E1Mode        = OTX_T1E1_MODE_T1;
  // SPM3 is no longer a supported load module. The user can over ride
  // the setting, or the default module for the ASM will be selected
  // in DemoOpenPhysicalDevices().
  envP->eProgramModule   = OTX_DSP_SPM_OTXSPM3; 
  envP->pszDspOutFile    = NULL;
  envP->boardNo          = DEMO_DEFAULT_BOARD_NO;
  envP->nBoardType       = OTX_DEVICE_THOR_PCI;
  envP->coDsp            = 4;
  envP->eMfToneType      = OTX_TONE_MF_R1;
  envP->bUseDspDebugger  = OTX_FALSE;
  envP->bLiLoop          = OTX_FALSE;

  envP->bReceiveEqualizer = OTX_FALSE;
  envP->eTermMode         = OTX_T1E1_TERM_100_OHM;
  envP->bMonitorMode      = OTX_FALSE;
  envP->bAmplifier        = OTX_FALSE;
  envP->nRxThreshold      = DEMO_DEFAULT_RX_THRESHOLD; // in mV

  strcpy(envP->logFileName, envP->cwd);
  strcat(envP->logFileName, DEMO_DEFAULT_LOG_FILE);
  for (i = 1; i < argc; i++) {        // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } else {
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {   // Log file
          strcpy(envP->logFileName, argv[i] + 2);
        } else if (*(argv[i]+1) == '9') {   // DSP Out file (overrides the DSP Program module switch)
          envP->pszDspOutFile = argv[i] + 2;
        } else if (*(argv[i]+1) == 'Q') {   // Exit on error
          envP->bExitOnError = OTX_TRUE;
        } else if (*(argv[i]+1) == 'B') {   // Board number
          envP->boardNo = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'X') {   // DSP Debugger
          envP->bUseDspDebugger = OTX_TRUE;
        } else if (*(argv[i]+1) == '0') {   // Li Loop
          envP->bLiLoop = OTX_TRUE;
        } else if (*(argv[i]+1) == 'P') {   // Dsp Program module
          if (*(argv[i]+2) == '1') {          // OtxSpm1.out
            envP->eProgramModule = OTX_DSP_SPM_OTXSPM1;
          } else if (*(argv[i]+2) == '2') {   // OtxSpm2.out
            envP->eProgramModule = OTX_DSP_SPM_OTXSPM2;
          } else if (*(argv[i]+2) == 'X') {   // OtxSpm2Ex.out
            envP->eProgramModule = OTX_DSP_SPM_OTXSPM2EX;
          } else if (*(argv[i]+2) == 'N') {   // OtxSpm2c5510IntMem.out
            envP->eProgramModule = OTX_DSP_SPM_OTXSPM2_INTMEM;
          } else if (*(argv[i]+2) == 'M') {   // OtxSpm2Exc5510IntMem.out
            envP->eProgramModule = OTX_DSP_SPM_OTXSPM2EX_INTMEM;
          } else if (*(argv[i]+2) == 'E') {   // OtxSpm2c549ExtMet.out
            envP->eProgramModule = OTX_DSP_SPM_OTXSPM2_EXTMEM;
          } else if (*(argv[i]+2) == 'D') {   // OtxDtTx.out
            envP->eProgramModule = OTX_DSP_SPM_OTXDTTX;
          } else if (*(argv[i]+2) == 'H') {   // OtxHdlc.out
            envP->eProgramModule = OTX_DSP_SPM_OTXHDLC;
          } else {
            DemoLogf("Error: Invalid DSP Program Module command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
					  exit(1);
          }
        } else if (*(argv[i]+1) == 'D') {    // MF Detector Tone Type
          if (*(argv[i]+2) == '1') {         // R1
            envP->eMfToneType      = OTX_TONE_MF_R1;
          } else if (*(argv[i]+2) == 'F') {  // R2 Forward
            envP->eMfToneType      = OTX_TONE_MF_R2_FORWARD;
          } else if (*(argv[i]+2) == 'B') {  // R2 Backward
            envP->eMfToneType      = OTX_TONE_MF_R2_BACKWARD;
          } else {
            DemoLogf("Error: Invalid MF Tone Type command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
					  exit(1);
          }

        } else if (*(argv[i]+1) == 'T') {   // Board Type
          if ( *(argv[i]+2) == '2') {
            envP->nBoardType = OTX_DEVICE_THOR_PCI;
            envP->coLi  = 2;
            envP->coDsp = 4;
          } else if ( *(argv[i]+2) == 'T') {
            envP->nBoardType = OTX_DEVICE_THOR_2_PCI_PLUS;
            envP->coLi  = 2;
            envP->coDsp = 4;
          } else if ( *(argv[i]+2) == 'E') {
            envP->nBoardType = OTX_DEVICE_THOR_2_PCI_EXPRESS;
            envP->coLi  = 2;
            envP->coDsp = 4;
          } else if ( *(argv[i]+2) == '4') {
            envP->nBoardType = OTX_DEVICE_THOR_4_PCI_PLUS;
            envP->coLi  = 2;
            envP->coDsp = 4;
          } else if ( *(argv[i]+2) == '8') {
            envP->nBoardType = OTX_DEVICE_THOR_8_PCI;
            envP->coLi  = 2;
            envP->coDsp = 4;
          } else if ( *(argv[i]+2) == 'P') {
            envP->nBoardType = OTX_DEVICE_THOR_8_PCI_PLUS;
            envP->coLi  = 2;
            envP->coDsp = 4;
          } else if ( *(argv[i]+2) == 'M') {
            envP->nBoardType = OTX_DEVICE_THOR_PCMCIA;
            envP->coLi  = 2;
            envP->coDsp = 2;
          } else if ( *(argv[i]+2) == 'S') {
            envP->nBoardType  = OTX_DEVICE_SLEIPNIR_1_PCI;
            envP->coLi  = 1;
            envP->coDsp = 1;
          } else if ( *(argv[i]+2) == 'G') {
            envP->nBoardType  = OTX_DEVICE_GIMLE_16_PCI_PLUS;
            envP->coLi = 2;
            envP->coDsp = 4;
          } else if ( *(argv[i]+2) == 'C') {
            envP->nBoardType  = OTX_DEVICE_THOR_2_EXPRESSCARD;
            envP->coLi = 2;
            envP->coDsp = 4; // Not really. this card only has 1 DSP.
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }
        } else if (*(argv[i]+1) == 'M') {   // Li Mode
          if (*(argv[i]+2) == 'E') {   // E1 
            envP->eT1E1Mode    = OTX_T1E1_MODE_E1;
          } else if (*(argv[i]+2) == 'T') {   // T1
            envP->eT1E1Mode    = OTX_T1E1_MODE_T1;
          } else {
            DemoLogf("Error: Invalid Li Mode (E1/T1) command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
					  exit(1);
          }
        } else if (*(argv[i]+1) == 'C') {   // Companding Law
          if (*(argv[i]+2) == 'A') {   // A-law
            envP->law = OTX_DATA_FORMAT_COMPAND8_A_LAW;
          } else if (*(argv[i]+2) == 'U') {   // U-law
            envP->law = OTX_DATA_FORMAT_COMPAND8_U_LAW;
          } else {
            DemoLogf("Error: Invalid companding law command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
					  exit(1);
          }
        } else if (*(argv[i] + 1) == 'A') {
          // Use Front End Amplifier = ON
          envP->bAmplifier = OTX_TRUE;
        } else if (*(argv[i] + 1) == 'E') {
          // Use ReceiveEqualizer  = ON
          envP->bReceiveEqualizer = OTX_TRUE;
        } else if (*(argv[i] + 1) == '2') {
          // Use MonitorMode  = ON
          envP->bMonitorMode = OTX_TRUE;
        } else if (*(argv[i] + 1) == 'H') {
          // RxThreshold (in mV)
          envP->nRxThreshold = strtoul(argv[i] + 2, NULL, 0);
        } else if (*(argv[i] + 1) == 'Y') {
          // TermMode
          if ( *(argv[i]+2) == '1') {
            envP->eTermMode = OTX_T1E1_TERM_100_OHM;
          } else if ( *(argv[i]+2) == '7') {
            envP->eTermMode = OTX_T1E1_TERM_75_OHM;
          } else if ( *(argv[i]+2) == 'Z') {
            envP->eTermMode = OTX_T1E1_TERM_HIGHZ;
          } else if ( *(argv[i]+2) == 'S') {
            envP->eTermMode = OTX_T1E1_TERM_75_OHM_BALANCED;
          } else {
            DemoLogf("Error: Unknown Term Mode command option %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
				  exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
			  exit(1);
      }
    }
  }
}   


/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /B<boardNo>  : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /L<fileName> : Log file name\n");
  DemoLogf(" /T[2|T|E|4|8|P|M|G|C]: Board Type used for the demo, where\n");
  DemoLogf("                  2 = Use Thor-2-PCI Board with Vidar-5x4-ASM\n");
  DemoLogf("                  T = Use Thor-2-PCI-Plus Board with Vidar-ASM\n");
  DemoLogf("                  E = Use Thor-2-PCI-Express Board with Vidar-ASM\n");
  DemoLogf("                  4 = Use Thor-4-PCI-Plus Board with Vidar-ASM\n");
  DemoLogf("                  8 = Use Thor-8-PCI Board with Vidar-5x4-ASM\n");
  DemoLogf("                  P = Use Thor-8-PCI-Plus Board with Vidar-ASM \n");
  DemoLogf("                  M = Use Thor-2-PCMCIA Board\n");
  DemoLogf("                  G = Use Gimle-16-PCI-Plus with Vidar-ASM Board\n");
  DemoLogf("                  C = Use Thor-2-ExpressCard Board\n");
  DemoLogf(" /CA          : A-law\n");
  DemoLogf(" /CU          : u-law\n");
  DemoLogf(" /ME          : E1 Li mode\n");
  DemoLogf(" /MT          : T1 Li mode\n");
  DemoLogf(" /P[2|X|E|1|D|H|N|M] : DSP Program module, where\n");
  DemoLogf("                  2 = Use OtxSpm2 DSP module (default)\n");
  DemoLogf("                  X = Use OtxSpm2Ex DSP module\n");
  DemoLogf("                  E = Use OtxSpm2_ExtMem DSP module\n");
  DemoLogf("                  1 = Use OtxSpm1 DSP module\n");
  DemoLogf("                  D = Use OtxDtTx DSP module\n");
  DemoLogf("                  H = Use OtxHdlc DSP module\n");
  DemoLogf("                  N = Use OtxSpm2_IntMem DSP module\n");
  DemoLogf("                  M = Use OtxSpm2Ex_IntMem DSP module\n");
  DemoLogf(" /D1          : Use MF Detector R1 (default)\n");
  DemoLogf(" /DF          : Use MF Detector R2 Forward\n");
  DemoLogf(" /DB          : Use MF Detector R2 Backward\n");
  DemoLogf(" /A           : Front End Amplifier = ON (default is OFF)\n");
  DemoLogf(" /E           : Receive Equalizer = ON (longhaul) (default is OFF)\n");
  DemoLogf(" /2           : -20dB MonitorMode = ON (default is OFF)\n");
  DemoLogf(" /H           : Receive Input Threshold in mV (default is %d mV)\n", DEMO_DEFAULT_RX_THRESHOLD);
  DemoLogf(" /Y           : Termination Mode, where\n");
  DemoLogf("                  1 = 100/120 ohm balanced (default)\n");
  DemoLogf("                  7 = 75 ohm unbalanced\n");
  DemoLogf("                  Z = High-impedance\n");
  DemoLogf("                  S = 75 ohm balanced\n");
  DemoLogf(" /X           : DSP Debugger mode. Prompt to start DSP code from Debugger\n");
  DemoLogf(" /0           : Loop Lis (TX->RX)\n");
  DemoLogf(" /9<fileName> : DSP Out File (overrides the /P switch)\n");
  DemoLogf(" /Z<str>      : Command string (overrides keyboard input)\n");
  DemoLogf(" /Q           : Exit without prompting on errors\n");
}


/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(char *formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

}


static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {

    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf("ERROR(%#lX) \n", result);
    } else {
      OTX_RESULT otxRc;
      char msgStr[256];
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      eDeviceType = 999;    // Device type for hDevice 
      OTX_UINT32      nDeviceNumber = 999;  // Device number for hDevice
      
      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_TYPE, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType); 
      } 
      
      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_NO, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber); 
      }
      
      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(OTX_DRIVER_DEVICE_HANDLE, result, 256, &msgStr[0]);
      if (OTX_FAILURE(otxRc)) {
        strcpy(msgStr, "Cannot translate this result code.");
      }
      DemoLogf("ERROR(%#lX [%d.%d]) Dev(%X,%d,%d): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
    }

    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error 
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}


/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
OTX_INT32 DemoProcessUserCommand(char commandChar)
{
  static OTX_HANDLE hCurrentLi = 0;
  static OTX_INT32  nCurrentLi = 0;
  OTX_RESULT result;

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
    DemoLogf(" 0-1 Select active Li, this line is then used for subsequent commands\n");
    DemoLogf(" t   Send a HDLC frame on the current Li\n");
    DemoLogf(" e   Enable board interrupts\n");
    DemoLogf(" d   Disable board interrupts\n");
    DemoLogf(" a   Sending HDLC frames continuously\n");
    DemoLogf(" g/G Sending a string of DTMF digits on TS1/TS2 (from current Li)\n");
    if (g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2 || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2EX || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2_INTMEM || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2EX_INTMEM || 
        g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2_EXTMEM) {
      DemoLogf(" f/F Sending a string of MF digits on TS5/TS6 (from current Li)\n");
    }
    DemoLogf(" i/I Generate a generic tone on Li#%d (TS%d/TS%d) \n", nCurrentLi, 5, 6);
    DemoLogf(" j/J Send one DTMF tone on TS5/TS6 (from current Li)\n");
    if (g_demoEnv.eMfToneType == OTX_TONE_MF_R1) {
      DemoLogf(" k/K Send one R1 MF tone on TS5/TS6 (from current Li) \n");
    }
    if (g_demoEnv.eMfToneType == OTX_TONE_MF_R2_FORWARD) {
      DemoLogf(" l/L Send one R2 MF Forward tone on TS5/TS6 (from current Li) \n");
    }
    if (g_demoEnv.eMfToneType == OTX_TONE_MF_R2_BACKWARD) {
      DemoLogf(" m/M Send one R2 MF Backward tone on TS5/TS6 (from current Li) \n");
    }
    DemoLogf(" r   Disable/Enable DTMF Dialer devices\n");
    DemoLogf(" R   Reset DTMF Dialer device multiple times\n");
    if (g_demoEnv.nBoardType != OTX_DEVICE_THOR_PCMCIA) {
      DemoLogf(" >   Increment TS for codecs (current TS=%d)\n", g_nCodecTs);
      DemoLogf(" <   Decrement TS for codecs (current TS=%d)\n", g_nCodecTs);
    }
    break;

 case 'r':
    {
      OTX_INT32 nDialler;
      
      DemoLogf("Disable/Enable DTMF dialler[0-3]\n");
      for (nDialler = 0; nDialler < g_demoEnv.coDsp; nDialler++) {
        result = OtxDrvDisable(g_hLogicDevs.hDtmfDialler[nDialler],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hDtmfDialler[nDialler], result);
      }
      for (nDialler = 0; nDialler < g_demoEnv.coDsp; nDialler++) {
        result = OtxDrvEnable(g_hLogicDevs.hDtmfDialler[nDialler],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hDtmfDialler[nDialler], result);
      }
      DemoLogf("Done.");
    }
    break;

 case 'R':
    {
      OTX_INT32 nDialler;
      static OTX_INT32 nLoopTimes = 1;
      OTX_INT32 i;
      char strBuf[12];
      
      DemoLogf("Reseting DTMF dialler[0-3]\n");
      DemoLogf("How many times? (%d)", nLoopTimes);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        nLoopTimes = strtol(strBuf, NULL, 0);
      }

      for (i = 0; i < nLoopTimes; i++) {
        for (nDialler = 0; nDialler < g_demoEnv.coDsp; nDialler++) {
          result = OtxDrvDisable(g_hLogicDevs.hDtmfDialler[nDialler],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hDtmfDialler[nDialler], result);
        }
        for (nDialler = 0; nDialler < g_demoEnv.coDsp; nDialler++) {
          result = OtxDrvEnable(g_hLogicDevs.hDtmfDialler[nDialler],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hDtmfDialler[nDialler], result);
        }
      }
      DemoLogf("Done.");
    }
    break;

  case 'B':
    {
      OTX_INT32 nDsp;
      
      for (nDsp = 0; nDsp < g_demoEnv.coDsp; nDsp++) {
        DemoLogf("Init Bsp for Dsp#%d\n", nDsp);
        OtxDrvDbgDump(g_hPhysDevs.hDsp[nDsp], 1);
      }
    }
    break;

  case ':':
    {
      OTX_INT32 nDsp;
      
      for (nDsp = 0; nDsp < 1; nDsp++) {
        DemoLogf("Debug Dump for Dsp#%d\n", nDsp);
        OtxDrvDbgDump(g_hPhysDevs.hDsp[nDsp], 0);
      }
    }
    break;

  case 'g':
    {
      OTX_INT32 nDigits = strlen(g_sqNumber0);
      OTX_HANDLE hHandle = g_hLogicDevs.hDtmfDialler[nCurrentLi];

      DemoLogf("5 Sending digits on Li#%d TS%d '%s'\n", nCurrentLi, 1, g_sqNumber0);

      // Start Dialing DTMF number, Non-Blocking call, will receive Event when completed
      result = OtxToneDtmfDial(hHandle, (OTX_UINT8 *)g_sqNumber0, nDigits, DEMO_TASK_REF, OTX_TIME_INFINITY); 
      if (OTX_RESULT_CODE(result) == OTX_DTMF_DIALER_RES_BUSY) {
        DemoLogf("Dialer busy.\n");
      } else {
        DemoCheckRc(hHandle, result);
        DemoLogf("Done.\n");
      }

    }
    break;


  case 'f':
    {
      OTX_HANDLE hHandle = g_hLogicDevs.hMfDialler[nCurrentLi*2];
      char *sqDigitString;

      if (g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2 || 
          g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2EX || 
          g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2_INTMEM || 
          g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2EX_INTMEM || 
          g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2_EXTMEM) {
        
        switch (g_demoEnv.eMfToneType) {
        case OTX_TONE_MF_R1:
          sqDigitString = "0123456789abcde";
          break;
        case OTX_TONE_MF_R2_FORWARD:
        case OTX_TONE_MF_R2_BACKWARD:
          sqDigitString = "123456789abcdef";
          break;
        default:
          DemoLogf("Unsupported value for g_demoEnv.eMfToneType (%d) \n", g_demoEnv.eMfToneType);
          sqDigitString = "1";
          break;
        }
        
        
        DemoLogf("Sending digits on Li#%d TS%d '%s'\n", nCurrentLi, 1, sqDigitString);
        
        // Start Dialing MF digit string, Non-Blocking call, will receive Event when completed
        result = OtxToneMfDial(hHandle, (OTX_UINT8 *)sqDigitString, strlen(sqDigitString), DEMO_TASK_REF, OTX_TIME_INFINITY);
        if (OTX_RESULT_CODE(result) == OTX_MF_DIALER_RES_BUSY) {
          DemoLogf("Dialer busy.\n");
        } else {
          DemoCheckRc(hHandle, result);
          DemoLogf("Done.\n");
        }
        
      }

    }
    break;


  case 'i':
    {
      DemoLogf("Generate a generic tone on Li#%d TS%d \n", nCurrentLi, 5);

      DemoSendGenericSineWave(nCurrentLi*2);

      DemoLogf("Done.\n");
    }
    break;

  
  case 'I':
    {
      DemoLogf("Generate a generic tone on Li#%d TS%d \n", nCurrentLi, 6);

      DemoSendGenericSineWave(nCurrentLi*2 + 1);

      DemoLogf("Done.\n");
    }
    break;

  


  case 'j':
    {
      DemoLogf("Sending SineWaveTone on Li#%d TS%d \n", nCurrentLi, 5);

      DemoSendSineWaveDtmfTone(nCurrentLi*2);

      DemoLogf("Done.\n");
    }
    break;

  
  case 'J':
    {
      DemoLogf("Sending SineWaveTone on Li#%d TS%d \n", nCurrentLi, 6);

      DemoSendSineWaveDtmfTone(nCurrentLi*2 + 1);

      DemoLogf("Done.\n");
    }
    break;

  case 'k':
    {
      DemoLogf("Sending SineWaveTone on Li#%d TS%d \n", nCurrentLi, 5);

      DemoSendSineWaveMfR1T1Tone(nCurrentLi*2);

      DemoLogf("Done.\n");
    }
    break;

  
  case 'K':
    {
      DemoLogf("Sending SineWaveTone on Li#%d TS%d \n", nCurrentLi, 6);

      DemoSendSineWaveMfR1T1Tone(nCurrentLi*2 + 1);

      DemoLogf("Done.\n");
    }
    break;

  case 'l':
    {
      DemoLogf("Sending SineWaveTone on Li#%d TS%d \n", nCurrentLi, 5);

      DemoSendSineWaveMfR2E1ForwardTone(nCurrentLi*2);

      DemoLogf("Done.\n");
    }
    break;

  
  case 'L':
    {
      DemoLogf("Sending SineWaveTone on Li#%d TS%d \n", nCurrentLi, 6);

      DemoSendSineWaveMfR2E1ForwardTone(nCurrentLi*2 + 1);

      DemoLogf("Done.\n");
    }
    break;

  case 'm':
    {
      DemoLogf("Sending SineWaveTone on Li#%d TS%d \n", nCurrentLi, 5);

      DemoSendSineWaveMfR2E1BackwardTone(nCurrentLi*2);

      DemoLogf("Done.\n");
    }
    break;

  
  case 'M':
    {
      DemoLogf("Sending SineWaveTone on Li#%d TS%d \n", nCurrentLi, 6);

      DemoSendSineWaveMfR2E1BackwardTone(nCurrentLi*2 + 1);

      DemoLogf("Done.\n");
    }
    break;

  case 'x':
    {
      OTX_RESULT result;
      OtxDeviceStateE  eDevState;
      OTX_HANDLE hHandle = g_hLogicDevs.hDtmfDialler[nCurrentLi*2];
      
      result = OtxDrvGetState(hHandle, &eDevState); DemoCheckRc(hHandle, result);
      if (eDevState == OTX_DEVICE_WORKING) {
        DemoLogf("Disabling sending device on Li#%d TS%d \n" , nCurrentLi, 1);
        result = OtxDrvDisable(hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(hHandle, result);
      } else {
        DemoLogf("Enabling sending device on Li#%d TS%d \n" , nCurrentLi, 1);
        result = OtxDrvEnable(hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(hHandle, result);
      }
      
      DemoLogf("Done.\n");
    }
    break;

  case 'X':
    {
      OTX_RESULT result;
      OtxDeviceStateE  eDevState;
      OTX_HANDLE hHandle = g_hLogicDevs.hDtmfDialler[nCurrentLi*2 + 1];
      
      result = OtxDrvGetState(hHandle, &eDevState); DemoCheckRc(hHandle, result);
      if (eDevState == OTX_DEVICE_WORKING) {
        DemoLogf("Disabling sending device on Li#%d TS%d \n" , nCurrentLi, 2);
        result = OtxDrvDisable(hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(hHandle, result);
      } else {
        DemoLogf("Enabling sending device on Li#%d TS%d \n" , nCurrentLi, 2);
        result = OtxDrvEnable(hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(hHandle, result);
      }
      
      DemoLogf("Done.\n");
    }
    break;



  case 'e':
    DemoLogf("Enabling board interrupts\n");
    OtxDrvEnable(g_hPhysDevs.hThorController0, DEMO_TASK_REF, OTX_TIME_INFINITY);
    break;

  case 'd':
    DemoLogf("Disabling board interrupts\n");
    OtxDrvDisable(g_hPhysDevs.hThorController0, DEMO_TASK_REF, OTX_TIME_INFINITY);
    break;

  case 'z':
    {
      OtxT1E1TermModeE eTermMode;

      DemoLogf("Set Line Termination mode:\n");
      DemoLogf("0 = OTX_T1E1_TERM_120_OHM\n");
      DemoLogf("1 = OTX_T1E1_TERM_100_OHM\n");
      DemoLogf("2 = OTX_T1E1_TERM_75_OHM\n");
      DemoLogf("3 = OTX_T1E1_TERM_HIGHZ\n");
      switch (OsaConsoleGetCh() - '0') {
      case 0:
        eTermMode = OTX_T1E1_TERM_120_OHM;
        break;
      case 1:
        eTermMode = OTX_T1E1_TERM_100_OHM;
        break;
      case 2:
        eTermMode = OTX_T1E1_TERM_75_OHM;
        break;
      case 3:
        eTermMode = OTX_T1E1_TERM_HIGHZ;
        break;
      default:
        DemoLogf("Invalid entry. Setting Line Termination to OTX_T1E1_TERM_100_OHM\n");
        eTermMode = OTX_T1E1_TERM_100_OHM;
        break;
      }

      result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[nCurrentLi], eTermMode); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
    }
    break;


  case 't':
    {
      OTX_UINT8 arMsg[64];
      OTX_INT32 i;
      OTX_INT32 nMsgLen = 10;
      OTX_HANDLE hHandle;

      hHandle = g_hLogicDevs.hHdlcSender[nCurrentLi];

      for (i = 0; i < nMsgLen; i++) {
        arMsg[i] = (OTX_UINT8) i;
      }

      DemoLogf("Sending message on Li #%d ...", nCurrentLi);

      // Send HDLC frame
      result = OtxHdlcSendData(hHandle, arMsg, nMsgLen, DEMO_TASK_REF, OTX_TIME_INFINITY); DemoCheckRc(hHandle, result);

      DemoLogf("Done.\n");
    }
    break;

#if 1
  case 'T':
    {
      OTX_UINT8 *pBadMsg = (OTX_UINT8 *)"\x00\x01\x26\x3a\x4f\x02\x02\x68\x4d\x08\x03\x82\x90\x00";
      OTX_INT32 nMsgLen = 14;
      //OTX_UINT8 *pBadMsg = "\x04\x09\x7F";
      //OTX_INT32 nMsgLen = 3;
      OTX_HANDLE hHandle;

      hHandle = g_hLogicDevs.hHdlcSender[nCurrentLi];

      DemoLogf("Sending message on Li #%d ...", nCurrentLi);

      result = OtxHdlcSendData(hHandle, pBadMsg, nMsgLen, DEMO_TASK_REF, OTX_TIME_INFINITY); DemoCheckRc(hHandle, result);

      DemoLogf("Done.\n");
    }
    break;
#endif

  case '0':
    // Select POTS Line 0 as current
    nCurrentLi = 0;
    hCurrentLi = g_hPhysDevs.hLi[nCurrentLi];
    DemoLogf("LI#%d Selected ....\n", nCurrentLi);
    break;

  case '1':
    // Select POTS Line 1 as current
    nCurrentLi = 1;
    hCurrentLi = g_hPhysDevs.hLi[nCurrentLi];
    DemoLogf("LI#%d Selected ....\n", nCurrentLi);
    break;

  case '>':
    {
      if (g_demoEnv.nBoardType != OTX_DEVICE_THOR_PCMCIA) {
        g_nCodecTs++;
        if (g_nCodecTs > 31) {
          g_nCodecTs = 0;
        }
        
        DemoConnectCodecSpeakerSource(g_nCodecTs);
          
        DemoLogf("All codecs are now connected to TS%d (DSP highways)\n", g_nCodecTs);
      } else {
        DemoLogf("Function not supported by this board type\n");
      }
    }
    break;

  case '<':
    {
      if (g_demoEnv.nBoardType != OTX_DEVICE_THOR_PCMCIA) {
        g_nCodecTs--;
        if (g_nCodecTs < 0) {
          g_nCodecTs = 31;
        }
        
        DemoConnectCodecSpeakerSource(g_nCodecTs);
        
        DemoLogf("All codecs are now connected to TS%d (DSP highways)\n", g_nCodecTs);
      } else {
        DemoLogf("Function not supported by this board type\n");
      }
    }
    break;

  case 'a':
    {
      g_bContinousHdlcSendingActive = !g_bContinousHdlcSendingActive;
      if (g_bContinousHdlcSendingActive) {
        OsaThreadStart(DemoSendHdlcContinously);
      } 
    }
    break;

  case '~':
    {
      char strBuf[12];
      OTX_INT32 nSleepTime;

      DemoLogf("Enter milliseconds to sleep :");
      OsaConsoleGetStr(strBuf);

      if (strBuf[0]) {
        nSleepTime = strtol(strBuf, NULL, 0);
        DemoLogf("Sleeping %d milliseconds\n", nSleepTime);
        OsaThreadSleep(nSleepTime);
      }
    }
    break;

  case 'q':
  case 'Q':
    return 1;
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}



/*******************************************************************************
*******************************************************************************/
int main(int argc, char *argv[])
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;

  OsaConsoleInit();

  DemoLogf("T2p5x4, %s, Odin TeleSystems Inc.\n", __DATE__);
  DemoLogf("DTMF, MF tone example program on OTX board.\n");
  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    OsaConsoleExit();
		exit(-1);
  }

  DemoLogf("Log file is %s\n", g_demoEnv.logFileName);
  DemoLogf("MF Tone Type is");
  switch (g_demoEnv.eMfToneType) {
  case OTX_TONE_MF_R1:
    DemoLogf("R1\n");
    break;
  case OTX_TONE_MF_R2_FORWARD:
    DemoLogf("R2 Forward\n");
    break;
  case OTX_TONE_MF_R2_BACKWARD:
    DemoLogf("R2 Backward\n");
    break;
  default:
    DemoLogf("\"Unknown (%d)\" \n", g_demoEnv.eMfToneType);
    break;
  }

  // Query Library revision
  {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_REV_STR, szVersionInfo, 255); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, OTX_ATTR_LIB_TYPE_STATIC, &bIsStaticLinkedLibrary); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_FILE_NAME, szLibFileName, 64); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    
    DemoLogf("Library version %s (filename %s, %s linked)\n", szVersionInfo, szLibFileName, bIsStaticLinkedLibrary ? "static":"dynamic");
  }


  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib(); 
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  default:
    DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    break;
  }

  // Create Event Queues for Driver --> Application Communications
  DemoLogf("Creating event queue...\n");
  DemoCreateEventQueue();
  DemoLogf("Created event queue\n");

  // Connect (get handles) to the physical devices on the Thor-2-PCI board
  DemoLogf("Opening physical devices...\n");
  DemoOpenPhysicalDevices();
  DemoLogf("Opened physical devices\n");

  // Configure E1/T1 devices
  DemoLogf("Configuring E1/T1 devices...\n");
  DemoConfigureE1T1Devices();
  DemoLogf("Configured E1/T1 devices\n");

  // Enable all physical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("Enabling physical devices...\n");
  DemoEnablePhysicalDevices();

  // Set Li0 as the clock source
  if (g_hPhysDevs.hVidarAsmController0) {
    // Thor-2-PCI, Thor-8-PCI and Thor-8-PCI-Plus
    result = OtxBrdSetClocks(g_hPhysDevs.hThorController0, OTX_CLOCK_MODE_LOCAL, OTX_CLOCK_SOURCE_LOCAL_0); DemoCheckRc(g_hPhysDevs.hThorController0, result);
#if 1  // Temporary
    {
      OTX_INT32 nDsp;
      for (nDsp = 0; nDsp < g_demoEnv.coDsp; nDsp++) {
        DemoLogf("Init Bsp for Dsp#%d\n", nDsp);
        OtxDrvDbgDump(g_hPhysDevs.hDsp[nDsp], 1);
      }
    }
#endif
  } else {
    // Thor-2-PCMCIA
    result = OtxBrdSetClocks(g_hPhysDevs.hThorController0, OTX_CLOCK_MODE_NA, OTX_CLOCK_SOURCE_LOCAL_0); DemoCheckRc(g_hPhysDevs.hThorController0, result);
  }

#if 0 // For testing
  {
    OTX_INT32 i;
    
    for (i = 0; i < 3000; i++) {
      // Create and close the Logical devices needed by this application
      DemoLogf("Creating and closing logical devices [%d]...\n", i);
      DemoCreateLogicalDevices();
      DemoCloseLogicalDevices();
    }
  }
#endif

  // Create and get handles to the Logical devices needed by this application
  DemoLogf("Creating logical devices...\n");
  DemoCreateLogicalDevices();
  DemoLogf("Created logical devices\n");

  // Create and get handles to the Pipe devices needed by this application
  DemoConnectDevices();

  // Initialized the Connected Logical Devices
  DemoInitializeLogicalDevices();

  // Enable all logical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoEnableLogicalDevices();

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....
	OsaThreadStart(DemoIdleFunction);

  // Read Serial Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszSerialNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_SERIAL_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszSerialNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_SERIAL_NUMBER, pszSerialNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    DemoLogf("Board Serial Number: %s\n", pszSerialNumber);
    free(pszSerialNumber);
  }

  // Read Build Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszBrdBuildNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszBrdBuildNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUILD, pszBrdBuildNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    DemoLogf("Board Build Number: %s\n", pszBrdBuildNumber);
    free(pszBrdBuildNumber);
  }

  if (g_demoEnv.nBoardType != OTX_DEVICE_THOR_PCMCIA) {
    // Read CPLD Build Numbers from the board
    {
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      char *pszCpldBuildNumbers;
      
      result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_CPLD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      assert(eAttribType == OTX_ATTRIB_SZ);
      assert(bReadOnly == OTX_TRUE);
      pszCpldBuildNumbers = (char *) malloc(cbAttribValue * sizeof(char));
      result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_CPLD_BUILD, pszCpldBuildNumbers, cbAttribValue); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      DemoLogf("CPLD Build:\n%s", pszCpldBuildNumbers);
      free(pszCpldBuildNumbers);
    }
  }

  OsaThreadSleep(300); // To let DemoIdleFunction start up
  DemoLogf("\nPress 'h' for help\n");

  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
    exitFlag = DemoProcessUserCommand(commandChar);
  }

  // Clean up
  OtxDrvReleaseLib();
	OsaConsoleExit();

  return 0;
}

#if defined(_WIN32_WCE)
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPWSTR lpCmdLine, int nShowCmd) 
{
  // default command line parameters
  WCHAR wsExeName[OTX_MAX_PATH * sizeof(WCHAR)];
  static char exename[OTX_MAX_PATH];
  static char *cmdParams[] = { exename, "/TM", "/P2", NULL }; 
  char *pc;

  GetModuleFileName(NULL, wsExeName, sizeof(wsExeName));
  WideCharToMultiByte(CP_ACP, 0, &wsExeName[0], sizeof(wsExeName), exename, sizeof(exename), NULL, NULL); 
  OsaStrCpy(g_demoEnv.cwd, cmdParams[0]);
  pc = OsaStrRChr(g_demoEnv.cwd, '\\');
  if (pc != NULL) {
    *++pc = '\0';
  }

  return OsaWinCeConsoleMain(hInstance, hPrevInstance, lpCmdLine, nShowCmd,
              cmdParams, sizeof(cmdParams) / sizeof(cmdParams[0]), main);
}
#endif  // #if defined(_WIN32_WCE)
